/*
 * Copyright (C) 2015, Charles University in Prague.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package cz.cuni.mff.d3s.stdynpor.analysis.synchevent;

import java.util.Set;
import java.util.HashSet;
import java.util.TreeSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;

import com.ibm.wala.ipa.callgraph.AnalysisScope;
import com.ibm.wala.ipa.callgraph.AnalysisCache;
import com.ibm.wala.ipa.callgraph.CallGraph;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.ipa.callgraph.impl.Everywhere;
import com.ibm.wala.ipa.cfg.ExplodedInterproceduralCFG;
import com.ibm.wala.ipa.cfg.BasicBlockInContext;
import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.types.TypeReference;
import com.ibm.wala.ssa.SSAOptions;
import com.ibm.wala.ssa.IR;
import com.ibm.wala.ssa.DefUse;
import com.ibm.wala.ssa.SSAInstruction;
import com.ibm.wala.ssa.SSAInvokeInstruction;
import com.ibm.wala.ssa.SSAReturnInstruction;
import com.ibm.wala.ssa.SSAMonitorInstruction;
import com.ibm.wala.ssa.SSAFieldAccessInstruction;
import com.ibm.wala.ssa.SSAGetInstruction;
import com.ibm.wala.ssa.SSAPutInstruction;
import com.ibm.wala.ssa.SSALoadMetadataInstruction;
import com.ibm.wala.ssa.analysis.IExplodedBasicBlock;
import com.ibm.wala.ssa.analysis.ExplodedControlFlowGraph;
import com.ibm.wala.util.intset.BitVector;
import com.ibm.wala.util.intset.OrdinalSetMapping;
import com.ibm.wala.util.intset.MutableMapping;
import com.ibm.wala.util.intset.IntSet; 
import com.ibm.wala.util.intset.IntIterator;
import com.ibm.wala.util.graph.Graph;
import com.ibm.wala.util.graph.impl.GraphInverter;
import com.ibm.wala.fixpoint.BitVectorVariable;
import com.ibm.wala.fixpoint.UnaryOperator;
import com.ibm.wala.dataflow.graph.BitVectorSolver;
import com.ibm.wala.dataflow.graph.BitVectorFramework;
import com.ibm.wala.dataflow.graph.IKilldallFramework;
import com.ibm.wala.dataflow.graph.AbstractMeetOperator;
import com.ibm.wala.dataflow.graph.BitVectorKillGen;
import com.ibm.wala.dataflow.graph.BitVectorUnion;
import com.ibm.wala.dataflow.graph.BitVectorIdentity;
import com.ibm.wala.dataflow.graph.ITransferFunctionProvider;
import com.ibm.wala.analysis.typeInference.TypeInference;

import cz.cuni.mff.d3s.stdynpor.data.ProgramPoint;
import cz.cuni.mff.d3s.stdynpor.data.AllocationSite;
import cz.cuni.mff.d3s.stdynpor.data.FieldName;
import cz.cuni.mff.d3s.stdynpor.data.FieldAccess;
import cz.cuni.mff.d3s.stdynpor.data.FieldID;
import cz.cuni.mff.d3s.stdynpor.bytecode.ByteCodeUtils;
import cz.cuni.mff.d3s.stdynpor.wala.WALAUtils;
import cz.cuni.mff.d3s.stdynpor.wala.PointerAnalysisData;
import cz.cuni.mff.d3s.stdynpor.wala.TypeAnalysisData;
import cz.cuni.mff.d3s.stdynpor.wala.FieldAccessCodeInfo;
import cz.cuni.mff.d3s.stdynpor.wala.dataflow.BitVectorKillAll;
import cz.cuni.mff.d3s.stdynpor.wala.dataflow.BitVectorIntersection;

import cz.cuni.mff.d3s.stdynpor.Debug;


public class LockPatternThisAnalysis
{
	// for each method (signature), a map from a lock variable ("this" or a class object) to a set of program points inside a synchronized block over the lock
	protected static Map<String, Map<String, Set<ProgramPoint>>> mth2LockedProgPoints;

	// maps from program points to sets of field IDs
	// we consider object fields that are either accessed (read, write) through different variable than "this" or without lock over "this" (that actually means outside of synchronized blocks over "this")
	// we also consider accesses to static fields and synchronized blocks over their respective class objects
	// sets are negated (different variable than "this", without lock) for performance reasons
	protected static Map<ProgramPoint, Set<FieldID>> pp2FieldReadsNotThisLocked;	
	protected static Map<ProgramPoint, Set<FieldID>> pp2FieldWritesNotThisLocked;
	
	// mapping between field IDs and integer numbers (used in bitvectors)
	// it is used only for stage B (whole program inter-procedural analysis)
	private static OrdinalSetMapping<FieldID> progFieldsNumbering;
	
	
	static
	{
		mth2LockedProgPoints = new HashMap<String, Map<String, Set<ProgramPoint>>>();
		
		pp2FieldReadsNotThisLocked = new HashMap<ProgramPoint, Set<FieldID>>();
		pp2FieldWritesNotThisLocked = new HashMap<ProgramPoint, Set<FieldID>>();
		
		progFieldsNumbering = null;
	}

	
	public static void analyzeProgram(CallGraph clGraph, ExplodedInterproceduralCFG icfg) throws Exception
	{
		// stage A: for each method M find each program point P in M such that (1) a lock over "this" (or a class object) is held just before P and (2) the lock is both acquired and released in M
		
		// method signatures
		Set<String> processedMethods = new HashSet<String>();
		
		// analyze methods reachable in the call graph
		for (Iterator<CGNode> cgnIt = clGraph.iterator(); cgnIt.hasNext(); )
		{
			IMethod mth = cgnIt.next().getMethod();
		
			// we skip native methods and abstract methods
			if (mth.isNative() || mth.isAbstract()) continue;
			
			String mthSig = WALAUtils.getMethodSignature(mth);
			
			if (processedMethods.contains(mthSig)) continue;
			processedMethods.add(mthSig);
			
			IR ir = WALAUtils.cache.getIRFactory().makeIR(mth, Everywhere.EVERYWHERE, SSAOptions.defaultOptions());
			
			// create the control-flow graph of the given method
			ExplodedControlFlowGraph cfg = ExplodedControlFlowGraph.make(ir);
	
			MethodLocksOverThis mlot = new MethodLocksOverThis(mth, cfg, ir);
			BitVectorSolver<IExplodedBasicBlock> solverMLOT = mlot.analyze();
			
			// collect analysis results: all program points with the bit representing "this" set to '1'
			
			Map<String, Set<ProgramPoint>> lvar2ProgPoints = new HashMap<String, Set<ProgramPoint>>();
			
			for (IExplodedBasicBlock ebb : cfg) 
			{
				int insnPos = WALAUtils.getInsnBytecodePos(mth, ebb.getFirstInstructionIndex());
			
				ProgramPoint pp = new ProgramPoint(mthSig, insnPos);
			
				IntSet out = solverMLOT.getOut(ebb).getValue();
				if (out != null)
				{
					for (IntIterator outIt = out.intIterator(); outIt.hasNext(); )
					{
						int lvarNum = outIt.next();
					
						String lvarStr = mlot.getLockVarStringForNum(lvarNum);
						
						Set<ProgramPoint> lockedPPs = lvar2ProgPoints.get(lvarStr);
						if (lockedPPs == null)
						{
							lockedPPs = new HashSet<ProgramPoint>();
							lvar2ProgPoints.put(lvarStr, lockedPPs);
						}
						
						lockedPPs.add(pp);
					}
				}
			}
			
			mth2LockedProgPoints.put(mthSig, lvar2ProgPoints);
		}
		
		// stage B: for each program point P find the set of fields that are in the future accessed only through "this" and inside synchronized blocks over "this" (except constructors)
		// implementation uses a reversed approach (may analysis, merge operator union, empty set as initial data-flow value) for performance reasons
		
		// create the backwards-oriented control-flow graph of the whole program
		Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG = GraphInverter.invert(icfg);
		
		// prepare common data structures
		
		createProgFieldsNumbering(bwICFG);
		
		// analyze field reads
		
		FieldAccessNotThisLocked fantlReads = new FieldAccessNotThisLocked(bwICFG, true, false);
		BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solverReadsFANTL = fantlReads.analyze(); 
	
		// collect analysis results
		for (BasicBlockInContext<IExplodedBasicBlock> bb : bwICFG) 
		{
			IExplodedBasicBlock ebb = bb.getDelegate();
			
			int insnPos = WALAUtils.getInsnBytecodePos(bb.getNode(), ebb.getFirstInstructionIndex());
		
			String fullMethodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
		
			ProgramPoint pp = new ProgramPoint(fullMethodSig, insnPos);

			Set<FieldID> fields = new HashSet<FieldID>();
		
			IntSet out = solverReadsFANTL.getOut(bb).getValue();
			if (out != null)
			{
				for (IntIterator outIt = out.intIterator(); outIt.hasNext(); )
				{
					int fieldNum = outIt.next();
				
					FieldID fid = progFieldsNumbering.getMappedObject(fieldNum);
					
					fields.add(fid);
				}
			}
		
			pp2FieldReadsNotThisLocked.put(pp, fields);
		}

		// analyze field writes
		
		FieldAccessNotThisLocked fantlWrites = new FieldAccessNotThisLocked(bwICFG, false, true);
		BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solverWritesFANTL = fantlWrites.analyze(); 
	
		// collect analysis results
		for (BasicBlockInContext<IExplodedBasicBlock> bb : bwICFG)
		{
			IExplodedBasicBlock ebb = bb.getDelegate();
			
			int insnPos = WALAUtils.getInsnBytecodePos(bb.getNode(), ebb.getFirstInstructionIndex());
		
			String fullMethodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
		
			ProgramPoint pp = new ProgramPoint(fullMethodSig, insnPos);

			Set<FieldID> fields = new HashSet<FieldID>();
		
			IntSet out = solverWritesFANTL.getOut(bb).getValue();
			if (out != null)
			{
				for (IntIterator outIt = out.intIterator(); outIt.hasNext(); )
				{
					int fieldNum = outIt.next();
				
					FieldID fid = progFieldsNumbering.getMappedObject(fieldNum);
					
					fields.add(fid);
				}
			}
		
			pp2FieldWritesNotThisLocked.put(pp, fields);
		}
	}
	
	public static Set<FieldID> getFieldReadsNotThisLocked(ProgramPoint pp)
	{
		return pp2FieldReadsNotThisLocked.get(pp);
	}

	public static Set<FieldID> getFieldWritesNotThisLocked(ProgramPoint pp)
	{
		return pp2FieldWritesNotThisLocked.get(pp);
	}
	
	public static void printLockedProgramPointsInEachMethod()
	{
		System.out.println("PROGRAM POINTS IN SYNCHRONIZED BLOCKS (THIS, CLASS NAME)");
		System.out.println("========================================================");
		
		// sort the set of method signatures
		Set<String> methodSigs = new TreeSet<String>();
		methodSigs.addAll(mth2LockedProgPoints.keySet());
		
		for (String mthSig : methodSigs)
		{
			if ( ! Debug.isWatchedEntity(mthSig) ) continue;
			
			System.out.println(mthSig);
			
			Map<String, Set<ProgramPoint>> lvar2ProgPoints = mth2LockedProgPoints.get(mthSig);
			
			for (String lockVar : lvar2ProgPoints.keySet())
			{
				System.out.println("\t" + lockVar);
				
				Set<ProgramPoint> lockedPPs = lvar2ProgPoints.get(lockVar);
			
				for (ProgramPoint pp : lockedPPs) 
				{
					System.out.println("\t\t" + pp.toString());
				}
			}
		}
		
		System.out.println("");
	}
	
	public static void printFieldAccessesOnThisLocked(Map<ProgramPoint, Set<FieldID>> pp2FutureFieldReads, Map<ProgramPoint, Set<FieldID>> pp2FutureFieldWrites)
	{
		System.out.println("FIELDS ACCESSED ON LOCK VARIABLE (THIS, CLASS NAME) INSIDE SYNCHRONIZED BLOCKS");
		System.out.println("==============================================================================");
		
		// both maps from program points to field names have the same keys
		Set<ProgramPoint> progPoints = new TreeSet<ProgramPoint>();
		progPoints.addAll(pp2FieldReadsNotThisLocked.keySet());
		
		for (ProgramPoint pp : progPoints)
		{
			if ( ! Debug.isWatchedEntity(pp.methodSig) ) continue;
			
			Set<FieldID> fieldsNotRead = pp2FieldReadsNotThisLocked.get(pp);
			Set<FieldID> fieldsNotWritten = pp2FieldWritesNotThisLocked.get(pp);

			Set<FieldID> futureFieldReads = pp2FutureFieldReads.get(pp);
			Set<FieldID> futureFieldWrites = pp2FutureFieldWrites.get(pp);

			System.out.println(pp.methodSig + ":" + pp.insnPos);
			
			System.out.println("\t---- [fields read] ----");	
			for (FieldID fid : futureFieldReads)
			{
				if ( ! Debug.isWatchedEntity(fid.tgtObjectID.toString()) ) continue;
				
				if ( ! fieldsNotRead.contains(fid) ) System.out.println("\t" + fid.toString());
			}
			
			System.out.println("\t---- [fields written] ----");	
			for (FieldID fid : futureFieldWrites)
			{
				if ( ! Debug.isWatchedEntity(fid.tgtObjectID.toString()) ) continue;
				
				if ( ! fieldsNotWritten.contains(fid) ) System.out.println("\t" + fid.toString());
			}
		}
		
		System.out.println("");
	}
	
	private static void createProgFieldsNumbering(Graph<BasicBlockInContext<IExplodedBasicBlock>> icfg)
	{
		progFieldsNumbering = new MutableMapping<FieldID>(new FieldID[1]);
		
		for (BasicBlockInContext<IExplodedBasicBlock> bb : icfg)
		{
			CGNode mthNode = bb.getNode();
				
			IR methodIR = mthNode.getIR();
				
			if (methodIR == null) continue;
			
			String methodSig = WALAUtils.getMethodSignature(mthNode.getMethod());

			int firstInsnIdx = bb.getFirstInstructionIndex();
			int lastInsnIdx = bb.getLastInstructionIndex();
   
			// basic block without instructions
			if ((firstInsnIdx < 0) || (lastInsnIdx < 0)) continue;

			SSAInstruction[] instructions = methodIR.getInstructions();

			for (int insnIndex = firstInsnIdx; insnIndex <= lastInsnIdx; insnIndex++) 
			{
				SSAInstruction insn = instructions[insnIndex];
			
				if (insn instanceof SSAFieldAccessInstruction) 
				{
					SSAFieldAccessInstruction faInsn = (SSAFieldAccessInstruction) insn;
						
					try
					{
						Set<FieldID> fieldIDs = FieldAccessCodeInfo.getFieldsForInsn(mthNode, faInsn, insnIndex);
						for (FieldID fid : fieldIDs) progFieldsNumbering.add(fid);
					}
					catch (Exception ex) { ex.printStackTrace(); }						
				}
			}
		}
	}
	
	
	static class MethodLocksOverThis
	{
		private IMethod mth;
		
		private Graph<IExplodedBasicBlock> mthCFG;
		
		private IR mthIR;
		
		private DefUse mthDU;
		
		private TypeInference mthVarTypeInfo;

		protected OrdinalSetMapping<String> mthLockVarsNumbering;
		
		// all bits set to 1
		private BitVectorVariable fullBV;
		
		public MethodLocksOverThis(IMethod mth, Graph<IExplodedBasicBlock> mthCFG, IR mthIR)
		{
			this.mth = mth;
			this.mthCFG = mthCFG;
			this.mthIR = mthIR;			
			
			this.mthDU = new DefUse(mthIR);
			
			String mthSig = WALAUtils.getMethodSignature(mth);
			
			this.mthVarTypeInfo = TypeAnalysisData.getMethodTypeInfo(mthSig, mthIR);
			
			createMethodLockVariablesNumbering();
			
			fullBV = new BitVectorVariable();
			for (int i = 1; i <= mthLockVarsNumbering.getSize(); i++) fullBV.set(i);  
		}
	
		private void createMethodLockVariablesNumbering()
		{
			mthLockVarsNumbering = new MutableMapping<String>(new String[1]);
			
			mthLockVarsNumbering.add("this");
			
			try
			{
				// process static synchronized methods
				
				if (mth.isStatic() && mth.isSynchronized())
				{
					// get class name
					mthLockVarsNumbering.add(WALAUtils.getDeclaringClassNameForMethod(mth));	
				}			
				
				SSAInstruction[] instructions = mthIR.getInstructions();
				for (int i = 0; i < instructions.length; i++) 
				{
					SSAInstruction insn = instructions[i];
					
					// process entry to "synchronized (<class name>.class)"
					
					if (insn instanceof SSAMonitorInstruction)
					{
						SSAMonitorInstruction monInsn = (SSAMonitorInstruction) insn;					
						
						TypeReference tgtObjTypeRef = mthVarTypeInfo.getType(monInsn.getRef()).getTypeReference(); 
						String tgtObjClassName = WALAUtils.getClassName(tgtObjTypeRef);
						
						if (tgtObjClassName.equals("java.lang.Class"))
						{
							SSAInstruction tgtObjDefInsn = mthDU.getDef(monInsn.getRef());
							
							// def insn for monInsn.getRef() should be ldc_w
							if (tgtObjDefInsn instanceof SSALoadMetadataInstruction)
							{
								SSALoadMetadataInstruction metaInsn = (SSALoadMetadataInstruction) tgtObjDefInsn;
								
								// get class name from the constant pool
								String lvarClassName = WALAUtils.getClassName((TypeReference) metaInsn.getToken());
								
								mthLockVarsNumbering.add(lvarClassName);
							}
						}
					}		
				}
			}
			catch (Exception ex) { ex.printStackTrace(); }
		}
		
		protected String getLockVarStringForNum(int lvarNum)
		{
			return mthLockVarsNumbering.getMappedObject(lvarNum);
		}
		
		class TransferFunctionsMLOT implements ITransferFunctionProvider<IExplodedBasicBlock, BitVectorVariable> 
		{
			public AbstractMeetOperator<BitVectorVariable> getMeetOperator() 
			{
				return BitVectorIntersection.getInstance();
			}
	
			public UnaryOperator<BitVectorVariable> getEdgeTransferFunction(IExplodedBasicBlock src, IExplodedBasicBlock dst) 
			{
				return BitVectorIdentity.instance();
			}
			
			public UnaryOperator<BitVectorVariable> getNodeTransferFunction(IExplodedBasicBlock ebb) 
			{
				SSAInstruction insn = ebb.getInstruction();
				
				if (insn instanceof SSAMonitorInstruction)
				{
					SSAMonitorInstruction monInsn = (SSAMonitorInstruction) insn;
					
					BitVector gen = new BitVector();
					BitVector kill = new BitVector();

					// process entry to "synchronized (this)"
						
					// SSA value number for "this" is 1
					// static methods do not have "this"
					if ( ( ! mth.isStatic() ) && (monInsn.getRef() == 1) )
					{
						int lvarNum = mthLockVarsNumbering.getMappedIndex("this");
						
						if (monInsn.isMonitorEnter()) gen.set(lvarNum);
						else kill.set(lvarNum);
					}
					
					// process entry to "synchronized (<class name>.class)"
					
					try
					{
						TypeReference tgtObjTypeRef = mthVarTypeInfo.getType(monInsn.getRef()).getTypeReference(); 
						String tgtObjClassName = WALAUtils.getClassName(tgtObjTypeRef);
						
						if (tgtObjClassName.equals("java.lang.Class"))
						{
							SSAInstruction tgtObjDefInsn = mthDU.getDef(monInsn.getRef());
							
							// def insn for monInsn.getRef() should be ldc_w
							if (tgtObjDefInsn instanceof SSALoadMetadataInstruction)
							{
								SSALoadMetadataInstruction metaInsn = (SSALoadMetadataInstruction) tgtObjDefInsn;
								
								// get class name from the constant pool
								String lvarClassName = WALAUtils.getClassName((TypeReference) metaInsn.getToken());
								
								int lvarNum = mthLockVarsNumbering.getMappedIndex(lvarClassName);
							
								if (monInsn.isMonitorEnter()) gen.set(lvarNum);
								else kill.set(lvarNum);
							}
						}
					}
					catch (Exception ex) { ex.printStackTrace(); }
					
					return new BitVectorKillGen(kill, gen);
				}
				else
				{
					// identity function for all other instructions
					return BitVectorIdentity.instance();
				}
			}
			
			public boolean hasEdgeTransferFunctions() 
			{
				return true;
			}
			
			public boolean hasNodeTransferFunctions() 
			{
				return true;
			}
		}
		
		class BitVectorSolverMLOT extends BitVectorSolver<IExplodedBasicBlock>
		{
			public BitVectorSolverMLOT(IKilldallFramework<IExplodedBasicBlock, BitVectorVariable> problem) 
			{
				super(problem);
			}
			
			@Override
			protected BitVectorVariable makeNodeVariable(IExplodedBasicBlock n, boolean IN)
			{
				BitVectorVariable V = new BitVectorVariable();

				try
				{
					if (mth.isSynchronized())
					{
						if (mth.isStatic())
						{
							int lvarNum = mthLockVarsNumbering.getMappedIndex(WALAUtils.getDeclaringClassNameForMethod(mth));
							
							V.set(lvarNum);
						}
						else
						{
							int lvarNum = mthLockVarsNumbering.getMappedIndex("this");
							
							V.set(lvarNum);
						}
					}
				}
				catch (Exception ex) { ex.printStackTrace(); }
				
				return V;
			}
			
			@Override
			protected BitVectorVariable makeEdgeVariable(IExplodedBasicBlock src, IExplodedBasicBlock dst) 
			{
				BitVectorVariable V = new BitVectorVariable();

				try
				{
					if (mth.isSynchronized())
					{
						if (mth.isStatic())
						{
							int lvarNum = mthLockVarsNumbering.getMappedIndex(WALAUtils.getDeclaringClassNameForMethod(mth));
							
							V.set(lvarNum);
						}
						else
						{
							int lvarNum = mthLockVarsNumbering.getMappedIndex("this");
							
							V.set(lvarNum);
						}
					}
				}
				catch (Exception ex) { ex.printStackTrace(); }
				
				return V;
			}
		}
	 		
		public BitVectorSolver<IExplodedBasicBlock> analyze() 
		{
			BitVectorFramework<IExplodedBasicBlock, String> framework = new BitVectorFramework<IExplodedBasicBlock, String>(mthCFG, new TransferFunctionsMLOT(), mthLockVarsNumbering);
			
			BitVectorSolver<IExplodedBasicBlock> solver = new BitVectorSolverMLOT(framework);
			
			try
			{
				solver.solve(null);
			}
			catch (Exception e) {}
			
			return solver;
		} 
	}
	
	
	static class FieldAccessNotThisLocked
	{
		private Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG;
		
		private boolean considerReads;
		private boolean considerWrites;
				
		public FieldAccessNotThisLocked(Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG, boolean reads, boolean writes)
		{
			this.bwICFG = bwICFG;
			
			this.considerReads = reads;
			this.considerWrites = writes;
		}
		
		class TransferFunctionsFANTL implements ITransferFunctionProvider<BasicBlockInContext<IExplodedBasicBlock>, BitVectorVariable> 
		{
			public AbstractMeetOperator<BitVectorVariable> getMeetOperator() 
			{
				return BitVectorUnion.instance();
			}
			
			public UnaryOperator<BitVectorVariable> getNodeTransferFunction(BasicBlockInContext<IExplodedBasicBlock> bb) 
			{
				IExplodedBasicBlock ebb = bb.getDelegate();
				
				SSAInstruction insn = ebb.getInstruction();
				
				if (insn instanceof SSAFieldAccessInstruction) 
				{
					if ((considerReads && (insn instanceof SSAGetInstruction)) || (considerWrites && (insn instanceof SSAPutInstruction)))
					{
						SSAFieldAccessInstruction faInsn = (SSAFieldAccessInstruction) insn;
						
						try
						{
							Set<FieldID> tgtFieldIDs = FieldAccessCodeInfo.getFieldsForInsn(bb.getNode(), faInsn, ebb.getFirstInstructionIndex());
							
							String methodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
							int insnPos = WALAUtils.getInsnBytecodePos(bb.getNode(), ebb.getFirstInstructionIndex());
							ProgramPoint insnPP = new ProgramPoint(methodSig, insnPos);				
							
							// get results of the stage A for the current method
							Map<String, Set<ProgramPoint>> lvar2ProgPoints = mth2LockedProgPoints.get(methodSig);
							
							if ( ( ! faInsn.isStatic() ) && (faInsn.getRef() == 1) )
							{
								// field access via "this"
								
								if (bb.getNode().getMethod().isInit())
								{
									// we are inside object constructor
									
									// do nothing
									return BitVectorIdentity.instance();
								}
								else
								{
									// we are NOT inside object constructor (method "<init>")
									
									// check if the lock over "this" is held just before the field access
									
									Set<ProgramPoint> lockedPPs = lvar2ProgPoints.get("this");
									
									if ((lockedPPs != null) && lockedPPs.contains(insnPP))
									{
										// keep the current value of the bit representing target field
										return BitVectorIdentity.instance();
									}
								}
							}
							else if (faInsn.isStatic())
							{
								// check if the lock over the class object is held just before the field access
								
								String tgtFieldDeclClsName = WALAUtils.getDeclaringClassNameForField(faInsn.getDeclaredField());								
								
								Set<ProgramPoint> lockedPPs = lvar2ProgPoints.get(tgtFieldDeclClsName);
									
								if ((lockedPPs != null) && lockedPPs.contains(insnPP))
								{
									// keep the current value of the bit representing target field
									return BitVectorIdentity.instance();
								}
							}									
							
							// other cases for this instruction									
										
							BitVector gen = new BitVector();
							BitVector kill = new BitVector();
							
							for (FieldID tgtField : tgtFieldIDs)
							{
								// set the bit representing target field to '1'
								int fieldNum = progFieldsNumbering.getMappedIndex(tgtField);
								gen.set(fieldNum);
							}
							
							return new BitVectorKillGen(kill, gen);
						}						 
						catch (Exception ex) { ex.printStackTrace(); }
					}
				}
				
				// identity function for all other instructions
				return BitVectorIdentity.instance();
			}
			
			public UnaryOperator<BitVectorVariable> getEdgeTransferFunction(BasicBlockInContext<IExplodedBasicBlock> src, BasicBlockInContext<IExplodedBasicBlock> dst) 
			{
				// we do not propagate through return-to-exit edges 
				
				if (dst.getDelegate().isExitBlock()) 
				{
					return BitVectorKillAll.getInstance();
				}
				else
				{
					return BitVectorIdentity.instance();
				}
			}
			
			public boolean hasNodeTransferFunctions() 
			{
				return true;
			}
		
			public boolean hasEdgeTransferFunctions() 
			{
				return true;
			}
		}
		
		public BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> analyze() 
		{
			BitVectorFramework<BasicBlockInContext<IExplodedBasicBlock>, FieldID> framework = new BitVectorFramework<BasicBlockInContext<IExplodedBasicBlock>, FieldID>(bwICFG, new TransferFunctionsFANTL(), progFieldsNumbering);
			
			BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solver = new BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>>(framework);

			try
			{
				solver.solve(null);
			}
			catch (Exception ex) { ex.printStackTrace(); }
			
			return solver;
		}		
	}
}
