/*
 * Copyright (C) 2015, Charles University in Prague.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package cz.cuni.mff.d3s.stdynpor.analysis.synchevent;

import java.util.Set;
import java.util.HashSet;
import java.util.TreeSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;
import java.util.List;
import java.util.ArrayList;

import com.ibm.wala.ipa.callgraph.AnalysisScope;
import com.ibm.wala.ipa.callgraph.CallGraph;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.ipa.cfg.ExplodedInterproceduralCFG;
import com.ibm.wala.ipa.cfg.BasicBlockInContext;
import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.types.TypeReference;
import com.ibm.wala.ssa.IR;
import com.ibm.wala.ssa.DefUse;
import com.ibm.wala.ssa.SSAInstruction;
import com.ibm.wala.ssa.SSAInvokeInstruction;
import com.ibm.wala.ssa.SSAReturnInstruction;
import com.ibm.wala.ssa.SSAMonitorInstruction;
import com.ibm.wala.ssa.SSAFieldAccessInstruction;
import com.ibm.wala.ssa.SSAGetInstruction;
import com.ibm.wala.ssa.SSAPutInstruction;
import com.ibm.wala.ssa.SSALoadMetadataInstruction;
import com.ibm.wala.ssa.analysis.IExplodedBasicBlock;
import com.ibm.wala.util.intset.BitVector;
import com.ibm.wala.util.intset.OrdinalSetMapping;
import com.ibm.wala.util.intset.MutableMapping;
import com.ibm.wala.util.intset.IntSet; 
import com.ibm.wala.util.intset.IntIterator;
import com.ibm.wala.util.graph.Graph;
import com.ibm.wala.util.graph.impl.GraphInverter;
import com.ibm.wala.fixpoint.BitVectorVariable;
import com.ibm.wala.fixpoint.UnaryOperator;
import com.ibm.wala.dataflow.graph.BitVectorSolver;
import com.ibm.wala.dataflow.graph.BitVectorFramework;
import com.ibm.wala.dataflow.graph.AbstractMeetOperator;
import com.ibm.wala.dataflow.graph.BitVectorKillGen;
import com.ibm.wala.dataflow.graph.BitVectorUnion;
import com.ibm.wala.dataflow.graph.BitVectorIdentity;
import com.ibm.wala.dataflow.graph.ITransferFunctionProvider;

import cz.cuni.mff.d3s.stdynpor.data.ProgramPoint;
import cz.cuni.mff.d3s.stdynpor.data.AllocationSite;
import cz.cuni.mff.d3s.stdynpor.data.FieldID;
import cz.cuni.mff.d3s.stdynpor.data.SynchEventID;
import cz.cuni.mff.d3s.stdynpor.data.SynchEventType;
import cz.cuni.mff.d3s.stdynpor.data.FieldAccess;
import cz.cuni.mff.d3s.stdynpor.bytecode.ByteCodeUtils;
import cz.cuni.mff.d3s.stdynpor.wala.WALAUtils;
import cz.cuni.mff.d3s.stdynpor.wala.PointerAnalysisData;
import cz.cuni.mff.d3s.stdynpor.wala.SynchEventCodeInfo;
import cz.cuni.mff.d3s.stdynpor.wala.FieldAccessCodeInfo;
import cz.cuni.mff.d3s.stdynpor.wala.dataflow.BitVectorKillAll;
import cz.cuni.mff.d3s.stdynpor.wala.dataflow.BitVectorIntersection;
import cz.cuni.mff.d3s.stdynpor.wala.dataflow.BitVectorIntersectionFilterCR;
import cz.cuni.mff.d3s.stdynpor.util.TreeNode;
import cz.cuni.mff.d3s.stdynpor.util.Pair;
import cz.cuni.mff.d3s.stdynpor.Debug;


public class SynchronizationEventAnalysisBase
{
	// possible next field read accesses after the program point (exclusively)
	protected static Map<ProgramPoint, Set<FieldAccess>> pp2NextFieldReads;
	
	// possible next field write accesses after the program point (exclusively)
	protected static Map<ProgramPoint, Set<FieldAccess>> pp2NextFieldWrites;

	public static final FieldID EXIT_MARK_FIELD;
	
	// mapping between synchronization events and integer numbers (used in bitvectors)
	protected static OrdinalSetMapping<SynchEventID> eventsNumbering;

	// mapping between field accesses and integer numbers (used in bitvectors)
	protected static OrdinalSetMapping<FieldAccess> accessNumbering;

	// caches for field accesses between a given program point and the next access to a given field
	private static Map<Pair<ProgramPoint, FieldID>, TreeNode<FieldAccess>> cacheReadsBetween;
	private static Map<Pair<ProgramPoint, FieldID>, TreeNode<FieldAccess>> cacheWritesBetween;

	// custom merge operator intersection (with filtering) uses this to detect presence of relevant actions (synch event, field access) in nested method calls
	protected static final SynchEventID EXIT_MARK_SYNCHEVENT;

	// bit index used to represent the flag which identifies call-return edges during must-analysis
	private static int markCallRetBitPos;
	
	
	static
	{
		pp2NextFieldReads = null;
		pp2NextFieldWrites = null;
		
		EXIT_MARK_FIELD = new FieldID(new AllocationSite("<mark>", -1), "<exit>");
		
		eventsNumbering = null;
		
		accessNumbering = null;
		
		cacheReadsBetween = new HashMap<Pair<ProgramPoint, FieldID>, TreeNode<FieldAccess>>();
		cacheWritesBetween = new HashMap<Pair<ProgramPoint, FieldID>, TreeNode<FieldAccess>>();
		
		EXIT_MARK_SYNCHEVENT = new SynchEventID(new AllocationSite("<mark>", -1), SynchEventType.NONE);
	}
	
	
	static enum ApproximationLevel
	{
		MAY,
		MUST
	}

	
	protected static void createSynchEventNumbering(Graph<BasicBlockInContext<IExplodedBasicBlock>> icfg, CallGraph clGraph)
	{
		eventsNumbering = new MutableMapping<SynchEventID>(new SynchEventID[1]);
		
		for (BasicBlockInContext<IExplodedBasicBlock> bb : icfg)
		{
			CGNode mthNode = bb.getNode();
				
			IR methodIR = mthNode.getIR();
				
			if (methodIR == null) continue;
			
			String methodSig = WALAUtils.getMethodSignature(mthNode.getMethod());

			int firstInsnIdx = bb.getFirstInstructionIndex();
			int lastInsnIdx = bb.getLastInstructionIndex();
   
			// basic block without instructions
			if ((firstInsnIdx < 0) || (lastInsnIdx < 0)) continue;

			SSAInstruction[] instructions = methodIR.getInstructions();

			for (int idx = firstInsnIdx; idx <= lastInsnIdx; idx++) 
			{
				SSAInstruction insn = instructions[idx];
				
				try
				{
					int insnPos = WALAUtils.getInsnBytecodePos(mthNode, idx);
					
					// get synchronization events represented by the current instruction
					Set<SynchEventID> events = SynchEventCodeInfo.getSynchEventsForInsn(mthNode, methodIR, insn, clGraph);
					
					for (SynchEventID ev : events) eventsNumbering.add(ev);
				}
				catch (Exception ex) { ex.printStackTrace(); }
			}
		}
		
		// if the corresponding bit is set at a given program point then there was no field access and no synchronization event of relevant kinds between the point and exit node from the method
		eventsNumbering.add(EXIT_MARK_SYNCHEVENT);
		
		markCallRetBitPos = eventsNumbering.getMaximumIndex() + 1;
	}

	protected static SynchEventID getEventForNum(int eventNum)
	{
		return eventsNumbering.getMappedObject(eventNum);
	}
	
	protected static void computeNextFieldAccesses(CallGraph clGraph, Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG) throws Exception
	{
		// we already have the data from a previous invocation (of some analysis derived from this base)
		if ((pp2NextFieldReads != null) && (pp2NextFieldWrites != null)) return;
					
		pp2NextFieldReads = new HashMap<ProgramPoint, Set<FieldAccess>>();
		pp2NextFieldWrites = new HashMap<ProgramPoint, Set<FieldAccess>>();
				
		// stage B of analysis 3/6: for each program point P find all the possible nearest future field accesses after P (field names and code locations)
		
		// prepare common data structures
		
		createFieldAccessNumbering(bwICFG);
		
		// analyze field reads
		
		NextFieldAccesses nfaReads = new NextFieldAccesses(bwICFG, true, false);
		BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solverReadsNFA = nfaReads.analyze();
		
		// collect analysis results
		for (BasicBlockInContext<IExplodedBasicBlock> bb : bwICFG) 
		{
			IExplodedBasicBlock ebb = bb.getDelegate();
			
			int insnPos = WALAUtils.getInsnBytecodePos(bb.getNode(), ebb.getFirstInstructionIndex());
		
			String fullMethodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
		
			ProgramPoint pp = new ProgramPoint(fullMethodSig, insnPos);

			Set<FieldAccess> accesses = new HashSet<FieldAccess>();
		
			IntSet in = solverReadsNFA.getIn(bb).getValue();
			if (in != null)
			{
				for (IntIterator inIt = in.intIterator(); inIt.hasNext(); )
				{
					int accessNum = inIt.next();
				
					FieldAccess fa = getAccessForNum(accessNum);
				
					accesses.add(fa);
				}
			}
			
			pp2NextFieldReads.put(pp, accesses);
		}
		
		// analyze field writes
		
		NextFieldAccesses nfaWrites = new NextFieldAccesses(bwICFG, false, true);
		BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solverWritesNFA = nfaWrites.analyze();
		
		// collect analysis results
		for (BasicBlockInContext<IExplodedBasicBlock> bb : bwICFG) 
		{
			IExplodedBasicBlock ebb = bb.getDelegate();
			
			int insnPos = WALAUtils.getInsnBytecodePos(bb.getNode(), ebb.getFirstInstructionIndex());
		
			String fullMethodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
		
			ProgramPoint pp = new ProgramPoint(fullMethodSig, insnPos);

			Set<FieldAccess> accesses = new HashSet<FieldAccess>();
		
			IntSet in = solverWritesNFA.getIn(bb).getValue();
			if (in != null)
			{
				for (IntIterator inIt = in.intIterator(); inIt.hasNext(); )
				{
					int accessNum = inIt.next();
				
					FieldAccess fa = getAccessForNum(accessNum);
				
					accesses.add(fa);
				}
			}
			
			pp2NextFieldWrites.put(pp, accesses);
		}
	}
	
	public static TreeNode<FieldAccess> getFieldReadsBetween(ProgramPoint srcPP, FieldID dstField)
	{
		Pair<ProgramPoint, FieldID> ppfid = new Pair<ProgramPoint, FieldID>(srcPP, dstField);
		
		TreeNode<FieldAccess> reads = cacheReadsBetween.get(ppfid);
		
		if (reads == null)
		{
			reads = getFieldAccessesBetween(srcPP, dstField, pp2NextFieldReads);
			
			cacheReadsBetween.put(ppfid, reads);
		}
		
		return reads;
	}
	
	public static TreeNode<FieldAccess> getFieldWritesBetween(ProgramPoint srcPP, FieldID dstField)
	{
		Pair<ProgramPoint, FieldID> ppfid = new Pair<ProgramPoint, FieldID>(srcPP, dstField);
		
		TreeNode<FieldAccess> writes = cacheWritesBetween.get(ppfid);
		
		if (writes == null)
		{
			writes = getFieldAccessesBetween(srcPP, dstField, pp2NextFieldWrites);
			
			cacheWritesBetween.put(ppfid, writes);
		}
		
		return writes;
	}
	
	protected static TreeNode<FieldAccess> getFieldAccessesBetween(ProgramPoint srcPP, FieldID dstField, Map<ProgramPoint, Set<FieldAccess>> pp2NextAccesses)
	{
		Set<ProgramPoint> visitedPoints = new HashSet<ProgramPoint>();
		
		visitedPoints.add(srcPP);
		
		// root node will have a null item
		// its children correspond to next field accesses after 'srcPP'
		TreeNode<FieldAccess> accessesTreeRoot = new TreeNode<FieldAccess>(null);
		
		List<TreeNode<FieldAccess>> faTreeNodes = new ArrayList<TreeNode<FieldAccess>>();
		
		Set<FieldAccess> srcNextAccesses = pp2NextAccesses.get(srcPP);
		
		// explore graph of next field accesses recursively
		for (FieldAccess fa : srcNextAccesses)
		{
			TreeNode<FieldAccess> node = exploreFieldAccessGraphEdges(fa, dstField, visitedPoints, pp2NextAccesses);
			
			if (node == null) continue;
			
			if (node.item.targetField.equals(dstField)) continue;
			
			faTreeNodes.add(node);
		}
		
		accessesTreeRoot.children = faTreeNodes;
		
		return accessesTreeRoot;
	}
	
	protected static TreeNode<FieldAccess> exploreFieldAccessGraphEdges(FieldAccess curFA, FieldID dstField, Set<ProgramPoint> visitedPoints, Map<ProgramPoint, Set<FieldAccess>> pp2NextAccesses)
	{
		// we hit the "destination field" on this path
		if (curFA.targetField.equals(dstField)) return new TreeNode<FieldAccess>(curFA);
		
		// backtrack
		if (visitedPoints.contains(curFA.progPoint)) return null;
		
		visitedPoints.add(curFA.progPoint);
		
		// explore child nodes
		
		List<TreeNode<FieldAccess>> faTreeChildNodes = null;
		
		Set<FieldAccess> nextAccesses = pp2NextAccesses.get(curFA.progPoint);
		
		for (FieldAccess fa : nextAccesses)
		{
			TreeNode<FieldAccess> faChildNode =	exploreFieldAccessGraphEdges(fa, dstField, visitedPoints, pp2NextAccesses);
			
			if (faChildNode == null) continue;
			
			// we need to have the list only when some child node represents the "destination field" and we did not backtrack from all the children nodes 
			if (faTreeChildNodes == null) faTreeChildNodes = new ArrayList<TreeNode<FieldAccess>>();
				
			if ( ! faChildNode.item.targetField.equals(dstField) )
			{
				faTreeChildNodes.add(faChildNode);
			}
		}
		
		// we did not hit the "destination field" in the subtree and all nodes were already visited
		if (faTreeChildNodes == null) return null;
		
		TreeNode<FieldAccess> faTreeNode = new TreeNode<FieldAccess>(curFA);
			
		faTreeNode.children = faTreeChildNodes;			
		
		return faTreeNode;
	}
	
	public static void printNextFieldAccesses()
	{
		System.out.println("NEXT FIELD ACCESSES");
		System.out.println("===================");
		
		// both maps from program points to field accesses have the same keys
		Set<ProgramPoint> progPoints = new TreeSet<ProgramPoint>();
		progPoints.addAll(pp2NextFieldReads.keySet());
		
		for (ProgramPoint pp : progPoints)
		{
			if ( ! Debug.isWatchedEntity(pp.methodSig) ) continue;
			
			Set<FieldAccess> nextReads = pp2NextFieldReads.get(pp);
			Set<FieldAccess> nextWrites = pp2NextFieldWrites.get(pp);

			System.out.println(pp.methodSig + ":" + pp.insnPos);
			
			System.out.println("\t---- [next field reads] ----");
			for (FieldAccess fa : nextReads)
			{
				if ( ! Debug.isWatchedEntity(fa.progPoint.methodSig) ) continue;
				
				System.out.println("\t" + fa.toString());
			}
			
			System.out.println("\t---- [next field writes] ----");
			for (FieldAccess fa : nextWrites)
			{
				if ( ! Debug.isWatchedEntity(fa.progPoint.methodSig) ) continue;
				
				System.out.println("\t" + fa.toString());
			}
		}
		
		System.out.println("");
	}
	
	private static void createFieldAccessNumbering(Graph<BasicBlockInContext<IExplodedBasicBlock>> icfg)
	{
		accessNumbering = new MutableMapping<FieldAccess>(new FieldAccess[1]);
		
		for (BasicBlockInContext<IExplodedBasicBlock> bb : icfg)
		{
			CGNode mthNode = bb.getNode();
				
			IR methodIR = mthNode.getIR();
				
			if (methodIR == null) continue;
			
			if (bb.getDelegate().isExitBlock()) 
			{
				String methodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
				ProgramPoint exitPP = new ProgramPoint(methodSig, -1);
				
				FieldAccess exitMarkFA = new FieldAccess(EXIT_MARK_FIELD, exitPP);
				
				accessNumbering.add(exitMarkFA);
			}

			int firstInsnIdx = bb.getFirstInstructionIndex();
			int lastInsnIdx = bb.getLastInstructionIndex();
   
			// basic block without instructions
			if ((firstInsnIdx < 0) || (lastInsnIdx < 0)) continue;

			SSAInstruction[] instructions = methodIR.getInstructions();

			for (int idx = firstInsnIdx; idx <= lastInsnIdx; idx++) 
			{
				SSAInstruction insn = instructions[idx];
				
				try
				{
					if (insn instanceof SSAFieldAccessInstruction) 
					{
						SSAFieldAccessInstruction faInsn = (SSAFieldAccessInstruction) insn;
			
						// get field accessed performed by the current instruction
						Set<FieldAccess> accesses = FieldAccessCodeInfo.getFieldAccessesForInsn(mthNode, faInsn, idx);
						
						for (FieldAccess fa : accesses) accessNumbering.add(fa);
					}
				}
				catch (Exception ex) { ex.printStackTrace(); }
			}		
		}
	}
	
	protected static FieldAccess getAccessForNum(int accessNum)
	{
		return accessNumbering.getMappedObject(accessNum);
	}

	
	static class SynchEventsBeforeField
	{
		private Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG;

		private CallGraph clGraph;
		
		private ApproximationLevel approx;
		
		private boolean considerReads;
		private boolean considerWrites;
				
		public SynchEventsBeforeField(Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG, CallGraph clGraph, ApproximationLevel approx, boolean reads, boolean writes)
		{
			this.bwICFG = bwICFG;
			this.clGraph = clGraph;
			
			this.approx = approx;
			
			this.considerReads = reads;
			this.considerWrites = writes;
		}
				
		class TransferFunctionsSEBF implements ITransferFunctionProvider<BasicBlockInContext<IExplodedBasicBlock>, BitVectorVariable> 
		{
			public AbstractMeetOperator<BitVectorVariable> getMeetOperator() 
			{
				if (approx == ApproximationLevel.MUST)
				{
					int markEventNum = eventsNumbering.getMappedIndex(SynchronizationEventAnalysisBase.EXIT_MARK_SYNCHEVENT);
					
					return new BitVectorIntersectionFilterCR(markCallRetBitPos, markEventNum);
				}
				else
				{
					// MAY
					return BitVectorUnion.instance();
				}
			}

			public UnaryOperator<BitVectorVariable> getNodeTransferFunction(BasicBlockInContext<IExplodedBasicBlock> bb) 
			{
				IExplodedBasicBlock ebb = bb.getDelegate();
				
				SSAInstruction insn = ebb.getInstruction();

				if ((insn instanceof SSAInvokeInstruction) || (insn instanceof SSAReturnInstruction) || (insn instanceof SSAMonitorInstruction))
				{
					BitVector gen = new BitVector();
					BitVector kill = new BitVector();
				
					try
					{						
						Set<SynchEventID> events = SynchEventCodeInfo.getSynchEventsForInsn(bb.getNode(), bb.getNode().getIR(), insn, clGraph);
						
						for (SynchEventID ev : events)
						{
							int eventNum = eventsNumbering.getMappedIndex(ev);
							gen.set(eventNum);
						}
						
						// there was at least one synchronization event
						if ( ! events.isEmpty() )
						{
							int markEventNum = eventsNumbering.getMappedIndex(EXIT_MARK_SYNCHEVENT);
							kill.set(markEventNum);
						}
					}
					catch (Exception ex) { ex.printStackTrace(); }
					
					return new BitVectorKillGen(kill, gen);
				}
				else if (insn instanceof SSAFieldAccessInstruction) 
				{
					if ((considerReads && (insn instanceof SSAGetInstruction)) || (considerWrites && (insn instanceof SSAPutInstruction)))
					{					
						return BitVectorKillAll.getInstance();
					}
					else
					{
						// identity function for all other cases
						return BitVectorIdentity.instance();	
					}
				}
				else if (ebb.isExitBlock())
				{
					BitVector kill = new BitVector();
					BitVector gen = new BitVector();
					
					int markEventNum = eventsNumbering.getMappedIndex(EXIT_MARK_SYNCHEVENT);
					gen.set(markEventNum);
					
					return new BitVectorKillGen(kill, gen);
				}				
				else
				{
					// identity function for all other instructions
					return BitVectorIdentity.instance();
				}
			}
			
			public UnaryOperator<BitVectorVariable> getEdgeTransferFunction(BasicBlockInContext<IExplodedBasicBlock> src, BasicBlockInContext<IExplodedBasicBlock> dst) 
			{
				// we do not propagate through return-to-exit edges
				if (dst.getDelegate().isExitBlock()) 
				{					
					return BitVectorKillAll.getInstance();
				}
				
				if (approx == ApproximationLevel.MUST)
				{
					// detecting call-return edges: "dst" node is invoke and "src" is not entry node and "src" is not exit node
					if ( ( ! src.isEntryBlock() ) && ( ! src.isExitBlock() ) )
					{
						SSAInstruction insn = dst.getDelegate().getInstruction();
						
						if (insn instanceof SSAInvokeInstruction)
						{
							BitVector gen = new BitVector();
							BitVector kill = new BitVector();
							
							gen.set(markCallRetBitPos);
							
							return new BitVectorKillGen(kill, gen);
						}
					}
					
					// detecting exception edges: "src" is exit node and "dst" is not return instruction 
					if (src.isExitBlock())
					{
						SSAInstruction insn = dst.getDelegate().getInstruction();
						
						if ( ! (insn instanceof SSAReturnInstruction) )
						{
							BitVector gen = new BitVector(eventsNumbering.getMaximumIndex());
							
							// set all bits including the one representing EXIT_MARK_SYNCHEVENT
							// NOTE: we cannot use "gen.setAll()" here because it would also set few bits past the maximum defined index
							for (int i = 0; i <= eventsNumbering.getMaximumIndex(); i++) gen.set(i);
							
							gen.clear(markCallRetBitPos);
							
							BitVector kill = new BitVector();
							
							return new BitVectorKillGen(kill, gen);
						}					
					}					
				}
				
				// default
				return BitVectorIdentity.instance();
			}

			public boolean hasNodeTransferFunctions() 
			{
				return true;
			}
		
			public boolean hasEdgeTransferFunctions() 
			{
				return true;
			}
		}
	
		public BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> analyze() 
		{
			BitVectorFramework<BasicBlockInContext<IExplodedBasicBlock>, SynchEventID> framework = new BitVectorFramework<BasicBlockInContext<IExplodedBasicBlock>, SynchEventID>(bwICFG, new TransferFunctionsSEBF(), eventsNumbering);
			
			BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solver = new BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>>(framework);

			try
			{
				solver.solve(null);
			}
			catch (Exception ex) { ex.printStackTrace(); }
			
			return solver;
		}
	}
	
	
	static class NextFieldAccesses
	{
		private Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG;
		
		private boolean considerReads;
		private boolean considerWrites;
		
		public NextFieldAccesses(Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG, boolean reads, boolean writes)
		{
			this.bwICFG = bwICFG;
			
			this.considerReads = reads;
			this.considerWrites = writes;
		}
		
		class TransferFunctionsNFA implements ITransferFunctionProvider<BasicBlockInContext<IExplodedBasicBlock>, BitVectorVariable> 
		{
			public AbstractMeetOperator<BitVectorVariable> getMeetOperator() 
			{
				return BitVectorUnion.instance();
			}

			public UnaryOperator<BitVectorVariable> getNodeTransferFunction(BasicBlockInContext<IExplodedBasicBlock> bb) 
			{
				IExplodedBasicBlock ebb = bb.getDelegate();
				
				SSAInstruction insn = ebb.getInstruction();

				if (insn instanceof SSAFieldAccessInstruction) 
				{
					SSAFieldAccessInstruction faInsn = (SSAFieldAccessInstruction) insn;
					
					if ((considerReads && (faInsn instanceof SSAGetInstruction)) || (considerWrites && (faInsn instanceof SSAPutInstruction)))
					{			
						BitVector kill = new BitVector(accessNumbering.getSize()+1);
						kill.setAll();
						kill.clear(0);
						
						BitVector gen = new BitVector();
						
						try
						{
							Set<FieldAccess> accesses = FieldAccessCodeInfo.getFieldAccessesForInsn(bb.getNode(), faInsn, ebb.getFirstInstructionIndex());
							
							for (FieldAccess fa : accesses)
							{
								int accessNum = accessNumbering.getMappedIndex(fa);
								gen.set(accessNum);
								kill.clear(accessNum);
							}
						}
						catch (Exception ex) { ex.printStackTrace(); }
						
						return new BitVectorKillGen(kill, gen);
					}
					else
					{
						// identity function for all other cases
						return BitVectorIdentity.instance();
					}
				}
				else if (ebb.isExitBlock())
				{
					BitVector kill = new BitVector();
					BitVector gen = new BitVector();
					
					String methodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
					ProgramPoint exitPP = new ProgramPoint(methodSig, -1);
					
					FieldAccess exitMarkFA = new FieldAccess(EXIT_MARK_FIELD, exitPP);
					
					int faNum = accessNumbering.getMappedIndex(exitMarkFA);
					gen.set(faNum);
					
					return new BitVectorKillGen(kill, gen);
				}
				else
				{
					// identity function for all other instructions (nodes)
					return BitVectorIdentity.instance();
				}
			}
			
			public UnaryOperator<BitVectorVariable> getEdgeTransferFunction(BasicBlockInContext<IExplodedBasicBlock> src, BasicBlockInContext<IExplodedBasicBlock> dst) 
			{
				// we do not propagate through return-to-exit edges 
				
				if (dst.getDelegate().isExitBlock()) 
				{
					return BitVectorKillAll.getInstance();
				}
				else
				{
					return BitVectorIdentity.instance();
				}
			}

			public boolean hasNodeTransferFunctions() 
			{
				return true;
			}
		
			public boolean hasEdgeTransferFunctions() 
			{
				return true;
			}
		}
	
		public BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> analyze() 
		{
			BitVectorFramework<BasicBlockInContext<IExplodedBasicBlock>, FieldAccess> framework = new BitVectorFramework<BasicBlockInContext<IExplodedBasicBlock>, FieldAccess>(bwICFG, new TransferFunctionsNFA(), accessNumbering);
			
			BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solver = new BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>>(framework);

			try
			{
				solver.solve(null);
			}
			catch (Exception ex) { ex.printStackTrace(); }
			
			return solver;
		}
	}
	
}

