/*
 * Copyright (C) 2015, Charles University in Prague.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package cz.cuni.mff.d3s.stdynpor.analysis.synchevent;

import java.util.Set;
import java.util.HashSet;
import java.util.TreeSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;
import java.util.List;

import com.ibm.wala.ipa.callgraph.AnalysisScope;
import com.ibm.wala.ipa.callgraph.CallGraph;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.ipa.cfg.ExplodedInterproceduralCFG;
import com.ibm.wala.ipa.cfg.BasicBlockInContext;
import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.types.TypeReference;
import com.ibm.wala.ssa.IR;
import com.ibm.wala.ssa.SSAInstruction;
import com.ibm.wala.ssa.SSAInvokeInstruction;
import com.ibm.wala.ssa.SSAReturnInstruction;
import com.ibm.wala.ssa.SSAMonitorInstruction;
import com.ibm.wala.ssa.SSAFieldAccessInstruction;
import com.ibm.wala.ssa.SSAGetInstruction;
import com.ibm.wala.ssa.SSAPutInstruction;
import com.ibm.wala.ssa.analysis.IExplodedBasicBlock;
import com.ibm.wala.util.intset.BitVector;
import com.ibm.wala.util.intset.OrdinalSetMapping;
import com.ibm.wala.util.intset.MutableMapping;
import com.ibm.wala.util.intset.IntSet; 
import com.ibm.wala.util.intset.IntIterator;
import com.ibm.wala.util.graph.Graph;
import com.ibm.wala.util.graph.impl.GraphInverter;
import com.ibm.wala.fixpoint.BitVectorVariable;
import com.ibm.wala.fixpoint.UnaryOperator;
import com.ibm.wala.dataflow.graph.BitVectorSolver;
import com.ibm.wala.dataflow.graph.BitVectorFramework;
import com.ibm.wala.dataflow.graph.AbstractMeetOperator;
import com.ibm.wala.dataflow.graph.BitVectorKillGen;
import com.ibm.wala.dataflow.graph.BitVectorUnion;
import com.ibm.wala.dataflow.graph.BitVectorIdentity;
import com.ibm.wala.dataflow.graph.ITransferFunctionProvider;

import cz.cuni.mff.d3s.stdynpor.data.ProgramPoint;
import cz.cuni.mff.d3s.stdynpor.data.FieldID;
import cz.cuni.mff.d3s.stdynpor.data.SynchEventID;
import cz.cuni.mff.d3s.stdynpor.data.SynchEventType;
import cz.cuni.mff.d3s.stdynpor.data.FieldAccess;
import cz.cuni.mff.d3s.stdynpor.bytecode.ByteCodeUtils;
import cz.cuni.mff.d3s.stdynpor.wala.WALAUtils;
import cz.cuni.mff.d3s.stdynpor.wala.PointerAnalysisData;
import cz.cuni.mff.d3s.stdynpor.wala.dataflow.BitVectorKillAll;
import cz.cuni.mff.d3s.stdynpor.Debug;


public class SynchronizationEventMayAnalysis extends SynchronizationEventAnalysisBase
{
	// synchronization events between the program point (inclusively) and the next field read access
	protected static Map<ProgramPoint, Set<SynchEventID>> pp2SynchEventsNextRead;
	
	// synchronization events between the program point (inclusively) and the next field write access
	protected static Map<ProgramPoint, Set<SynchEventID>> pp2SynchEventsNextWrite;
	
	
	static
	{
		pp2SynchEventsNextRead = new HashMap<ProgramPoint, Set<SynchEventID>>();
		pp2SynchEventsNextWrite = new HashMap<ProgramPoint, Set<SynchEventID>>();
	}
	

	public static void analyzeProgram(CallGraph clGraph, ExplodedInterproceduralCFG icfg) throws Exception
	{
		// create the backwards-oriented control-flow graph of the whole program
		Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG = GraphInverter.invert(icfg);

		// stage A: for every program point P get the set of synchronization events between P and the nearest future field access
		
		// prepare common data structures
		
		createSynchEventNumbering(bwICFG, clGraph);

		// analyze field reads
		
		SynchEventsBeforeField sebfReads = new SynchEventsBeforeField(bwICFG, clGraph, ApproximationLevel.MAY, true, false);
		BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solverReadsSEBF = sebfReads.analyze(); 

		// collect analysis results
		for (BasicBlockInContext<IExplodedBasicBlock> bb : bwICFG) 
		{
			IExplodedBasicBlock ebb = bb.getDelegate();
			
			int insnPos = WALAUtils.getInsnBytecodePos(bb.getNode(), ebb.getFirstInstructionIndex());
		
			String fullMethodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
		
			ProgramPoint pp = new ProgramPoint(fullMethodSig, insnPos);

			Set<SynchEventID> events = new HashSet<SynchEventID>();
		
			IntSet res = null;
			if (WALAUtils.isFieldAccessInsn(bb.getNode(), ebb.getFirstInstructionIndex())) res = solverReadsSEBF.getIn(bb).getValue();
			else res = solverReadsSEBF.getOut(bb).getValue();
			if (res != null)
			{
				for (IntIterator resIt = res.intIterator(); resIt.hasNext(); )
				{
					int eventNum = resIt.next();
				
					SynchEventID sev = getEventForNum(eventNum);
				
					events.add(sev);
				}
			}
			
			pp2SynchEventsNextRead.put(pp, events);
		}
		
		// analyze field writes
		
		SynchEventsBeforeField sebfWrites = new SynchEventsBeforeField(bwICFG, clGraph, ApproximationLevel.MAY, false, true);
		BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solverWritesSEBF = sebfWrites.analyze(); 

		// collect analysis results
		for (BasicBlockInContext<IExplodedBasicBlock> bb : bwICFG) 
		{
			IExplodedBasicBlock ebb = bb.getDelegate();
			
			int insnPos = WALAUtils.getInsnBytecodePos(bb.getNode(), ebb.getFirstInstructionIndex());
		
			String fullMethodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
		
			ProgramPoint pp = new ProgramPoint(fullMethodSig, insnPos);

			Set<SynchEventID> events = new HashSet<SynchEventID>();
		
			IntSet res = null;
			if (WALAUtils.isFieldAccessInsn(bb.getNode(), ebb.getFirstInstructionIndex())) res = solverWritesSEBF.getIn(bb).getValue();
			else res = solverWritesSEBF.getOut(bb).getValue();
			if (res != null)
			{
				for (IntIterator resIt = res.intIterator(); resIt.hasNext(); )
				{
					int eventNum = resIt.next();
				
					SynchEventID sev = getEventForNum(eventNum);
				
					events.add(sev);
				}
			}
			
			pp2SynchEventsNextWrite.put(pp, events);
		}
		
		
		// stage B
		computeNextFieldAccesses(clGraph, bwICFG);
	}
	
	public static Set<SynchEventID> getSynchEventsBeforeNextRead(ProgramPoint pp)
	{
		return pp2SynchEventsNextRead.get(pp);
	}
	
	public static Set<SynchEventID> getSynchEventsBeforeNextWrite(ProgramPoint pp)
	{
		return pp2SynchEventsNextWrite.get(pp);
	}
	
	public static void printSynchEventsBeforeNextFieldAccess()
	{
		System.out.println("SYNCH EVENTS THAT MAY OCCUR BEFORE NEXT FIELD ACCESS");
		System.out.println("====================================================");
		
		// both maps from program points to synchronization events have the same keys
		Set<ProgramPoint> progPoints = new TreeSet<ProgramPoint>();
		progPoints.addAll(pp2SynchEventsNextRead.keySet());
		
		for (ProgramPoint pp : progPoints)
		{
			if ( ! Debug.isWatchedEntity(pp.methodSig) ) continue;
			
			Set<SynchEventID> eventsBeforeRead = pp2SynchEventsNextRead.get(pp);
			Set<SynchEventID> eventsBeforeWrite = pp2SynchEventsNextWrite.get(pp);

			System.out.println(pp.methodSig + ":" + pp.insnPos);
			
			System.out.println("\t---- [events before next read] ----");	
			for (SynchEventID ev : eventsBeforeRead)
			{
				if ( ! Debug.isWatchedEntity(ev.tgtObjectID.toString()) ) continue;
				
				System.out.println("\t" + ev.toString());
			}
			
			System.out.println("\t---- [events before next write] ----");	
			for (SynchEventID ev : eventsBeforeWrite)
			{
				if ( ! Debug.isWatchedEntity(ev.tgtObjectID.toString()) ) continue;
				
				System.out.println("\t" + ev.toString());
			}
		}
		
		System.out.println("");
	}
}
