/*
 * Copyright (C) 2015, Charles University in Prague.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package cz.cuni.mff.d3s.stdynpor.analysis.synchevent;

import java.util.Set;
import java.util.HashSet;
import java.util.TreeSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;
import java.util.List;
import java.util.EnumSet;

import com.ibm.wala.ipa.callgraph.AnalysisScope;
import com.ibm.wala.ipa.callgraph.CallGraph;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.ipa.cfg.ExplodedInterproceduralCFG;
import com.ibm.wala.ipa.cfg.BasicBlockInContext;
import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.types.TypeReference;
import com.ibm.wala.ssa.IR;
import com.ibm.wala.ssa.SSAInstruction;
import com.ibm.wala.ssa.SSAInvokeInstruction;
import com.ibm.wala.ssa.SSAReturnInstruction;
import com.ibm.wala.ssa.SSAMonitorInstruction;
import com.ibm.wala.ssa.SSAFieldAccessInstruction;
import com.ibm.wala.ssa.SSAGetInstruction;
import com.ibm.wala.ssa.SSAPutInstruction;
import com.ibm.wala.ssa.analysis.IExplodedBasicBlock;
import com.ibm.wala.util.intset.BitVector;
import com.ibm.wala.util.intset.OrdinalSetMapping;
import com.ibm.wala.util.intset.MutableMapping;
import com.ibm.wala.util.intset.IntSet; 
import com.ibm.wala.util.intset.IntIterator;
import com.ibm.wala.util.graph.Graph;
import com.ibm.wala.util.graph.impl.GraphInverter;
import com.ibm.wala.fixpoint.BitVectorVariable;
import com.ibm.wala.fixpoint.UnaryOperator;
import com.ibm.wala.dataflow.graph.BitVectorSolver;
import com.ibm.wala.dataflow.graph.BitVectorFramework;
import com.ibm.wala.dataflow.graph.IKilldallFramework;
import com.ibm.wala.dataflow.graph.AbstractMeetOperator;
import com.ibm.wala.dataflow.graph.BitVectorKillGen;
import com.ibm.wala.dataflow.graph.BitVectorUnion;
import com.ibm.wala.dataflow.graph.BitVectorIdentity;
import com.ibm.wala.dataflow.graph.ITransferFunctionProvider;

import cz.cuni.mff.d3s.stdynpor.data.ProgramPoint;
import cz.cuni.mff.d3s.stdynpor.data.AllocationSite;
import cz.cuni.mff.d3s.stdynpor.data.FieldID;
import cz.cuni.mff.d3s.stdynpor.data.SynchEventID;
import cz.cuni.mff.d3s.stdynpor.data.SynchEventType;
import cz.cuni.mff.d3s.stdynpor.data.FieldAccess;
import cz.cuni.mff.d3s.stdynpor.bytecode.ByteCodeUtils;
import cz.cuni.mff.d3s.stdynpor.wala.WALAUtils;
import cz.cuni.mff.d3s.stdynpor.wala.PointerAnalysisData;
import cz.cuni.mff.d3s.stdynpor.wala.FieldAccessCodeInfo;
import cz.cuni.mff.d3s.stdynpor.wala.SynchEventCodeInfo;
import cz.cuni.mff.d3s.stdynpor.wala.dataflow.BitVectorKillAll;
import cz.cuni.mff.d3s.stdynpor.wala.dataflow.BitVectorIntersection;
import cz.cuni.mff.d3s.stdynpor.wala.dataflow.BitVectorUnionFilterCR;
import cz.cuni.mff.d3s.stdynpor.Debug;


public class SynchronizationTypeMustAnalysis
{
	// a set of fields for each program point that satisfies this condition: for each field in the set there is a read/write access to the field between the program point and the nearest synchronization event of the given type (SET) on some control flow path starting in the program point
	protected static Map<ProgramPoint, Set<FieldID>> pp2FieldReadsBeforeSET;
	protected static Map<ProgramPoint, Set<FieldID>> pp2FieldWritesBeforeSET;
	
	protected static SynchEventType curEventType;
	
	// mapping between field IDs and integer numbers (used in bitvectors)
	private static OrdinalSetMapping<FieldID> fieldsNumbering;

	// custom merge operator union (with filtering) uses this to detect presence of relevant actions (synch event, field access) in nested method calls
	protected static final FieldID EXIT_MARK_FIELD;

	// bit index used to represent the flag which identifies call-return edges during must-analysis
	private static int markCallRetBitPos;

	
	static
	{
		pp2FieldReadsBeforeSET = new HashMap<ProgramPoint, Set<FieldID>>();
		pp2FieldWritesBeforeSET = new HashMap<ProgramPoint, Set<FieldID>>();
		
		curEventType = null;
		
		fieldsNumbering = null;
		
		EXIT_MARK_FIELD = new FieldID(new AllocationSite("<mark>", -1), "<exit>");
	}
	
	
	/**
	 * The parameter <code>evType</code> represents the type of synchronization events that we are interested in here (what synchronization events are we looking for).
	 */
	public static void analyzeProgram(CallGraph clGraph, ExplodedInterproceduralCFG icfg, SynchEventType evType) throws Exception
	{
		curEventType = evType;
		
		// create the backwards-oriented control-flow graph of the whole program
		Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG = GraphInverter.invert(icfg);

		// analysis definition: for every program point P get the set of fields that are accessed only after a synchronization event of the proper type		
		// implementation uses a reversed approach (may analysis, merge operator union, empty set as initial data-flow value) for performance reasons 
		
		// prepare common data structures
		
		createFieldsNumbering(bwICFG);
		
		// analyze field reads
		
		FieldsBeforeSynchEvent fbseReads = new FieldsBeforeSynchEvent(bwICFG, clGraph, true, false);
		BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solverReadsFBSE = fbseReads.analyze(); 
	
		// collect analysis results
		for (BasicBlockInContext<IExplodedBasicBlock> bb : bwICFG)
		{
			IExplodedBasicBlock ebb = bb.getDelegate();
			
			int insnPos = WALAUtils.getInsnBytecodePos(bb.getNode(), ebb.getFirstInstructionIndex());
		
			String fullMethodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
		
			ProgramPoint pp = new ProgramPoint(fullMethodSig, insnPos);
			
			Set<FieldID> fields = new HashSet<FieldID>();
					
			IntSet out = solverReadsFBSE.getOut(bb).getValue();
			if (out != null)
			{
				for (IntIterator outIt = out.intIterator(); outIt.hasNext(); )
				{
					int fieldNum = outIt.next();
				
					FieldID fid = fieldsNumbering.getMappedObject(fieldNum);
					
					fields.add(fid);
				}
			}
			
			pp2FieldReadsBeforeSET.put(pp, fields);
		}
		
		// analyze field writes
		
		FieldsBeforeSynchEvent fbseWrites = new FieldsBeforeSynchEvent(bwICFG, clGraph, false, true);
		BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solverWritesFBSE = fbseWrites.analyze(); 
	
		// collect analysis results
		for (BasicBlockInContext<IExplodedBasicBlock> bb : bwICFG)
		{
			IExplodedBasicBlock ebb = bb.getDelegate();
			
			int insnPos = WALAUtils.getInsnBytecodePos(bb.getNode(), ebb.getFirstInstructionIndex());
		
			String fullMethodSig = WALAUtils.getMethodSignature(bb.getNode().getMethod());
		
			ProgramPoint pp = new ProgramPoint(fullMethodSig, insnPos);
			
			Set<FieldID> fields = new HashSet<FieldID>();
		
			IntSet out = solverWritesFBSE.getOut(bb).getValue();
			if (out != null)
			{
				for (IntIterator outIt = out.intIterator(); outIt.hasNext(); )
				{
					int fieldNum = outIt.next();
				
					FieldID fid = fieldsNumbering.getMappedObject(fieldNum);
					
					fields.add(fid);
				}
			}
			
			pp2FieldWritesBeforeSET.put(pp, fields);
		}
	}
	
	public static boolean hasDataForSynchEventType(SynchEventType evType)
	{
		if (curEventType == null) return false;
		
		return (curEventType == evType);		
	}

	public static Set<FieldID> getFieldsReadBeforeSynchEvent(ProgramPoint pp)
	{
		return pp2FieldReadsBeforeSET.get(pp);
	}
	
	public static Set<FieldID> getFieldsWrittenBeforeSynchEvent(ProgramPoint pp)
	{
		return pp2FieldWritesBeforeSET.get(pp);
	}
	
	public static void printFieldsAfterSynchEvent(Map<ProgramPoint, Set<FieldID>> pp2FutureFieldReads, Map<ProgramPoint, Set<FieldID>> pp2FutureFieldWrites)
	{
		System.out.println("FIELDS ACCESSED ONLY AFTER SYNCH EVENT");
		System.out.println("======================================");
		
		// the key set is the same for both reads and writes
		Set<ProgramPoint> progPoints = new TreeSet<ProgramPoint>();
		progPoints.addAll(pp2FieldReadsBeforeSET.keySet());
		
		for (ProgramPoint pp : progPoints)
		{
			if ( ! Debug.isWatchedEntity(pp.methodSig) ) continue;
			
			Set<FieldID> allFieldsReadBeforeSET = pp2FieldReadsBeforeSET.get(pp);
			Set<FieldID> allFieldsWrittenBeforeSET = pp2FieldWritesBeforeSET.get(pp);
			
			Set<FieldID> futureFieldReads = pp2FutureFieldReads.get(pp);
			Set<FieldID> futureFieldWrites = pp2FutureFieldWrites.get(pp);

			System.out.println(pp.methodSig + ":" + pp.insnPos);
			
			System.out.println("\t---- [reads after event] ----");
			for (FieldID fid : futureFieldReads)
			{
				if ( ! Debug.isWatchedEntity(fid.tgtObjectID.toString()) ) continue;				
				
				if ( ! allFieldsReadBeforeSET.contains(fid) ) System.out.println("\t" + fid.toString());
			}
			
			System.out.println("\t---- [writes after event] ----");
			for (FieldID fid : futureFieldWrites)
			{
				if ( ! Debug.isWatchedEntity(fid.tgtObjectID.toString()) ) continue;				
				
				if ( ! allFieldsWrittenBeforeSET.contains(fid) ) System.out.println("\t" + fid.toString());
			}
		}
		
		System.out.println("");
	}
	
	private static void createFieldsNumbering(Graph<BasicBlockInContext<IExplodedBasicBlock>> icfg)
	{
		fieldsNumbering = new MutableMapping<FieldID>(new FieldID[1]);
		
		for (BasicBlockInContext<IExplodedBasicBlock> bb : icfg)
		{
			CGNode mthNode = bb.getNode();
			
			IR methodIR = mthNode.getIR();
			
			if (methodIR == null) continue;

			int firstInsnIdx = bb.getFirstInstructionIndex();
			int lastInsnIdx = bb.getLastInstructionIndex();
   
			// basic block without instructions
			if ((firstInsnIdx < 0) || (lastInsnIdx < 0)) continue;

			SSAInstruction[] instructions = methodIR.getInstructions();

			for (int insnIndex = firstInsnIdx; insnIndex <= lastInsnIdx; insnIndex++) 
			{
				SSAInstruction insn = instructions[insnIndex];

				if (insn instanceof SSAFieldAccessInstruction) 
				{
					SSAFieldAccessInstruction faInsn = (SSAFieldAccessInstruction) insn;
						
					try
					{
						Set<FieldID> fieldIDs = FieldAccessCodeInfo.getFieldsForInsn(mthNode, faInsn, insnIndex);
						for (FieldID fid : fieldIDs) fieldsNumbering.add(fid);
					}
					catch (Exception ex) { ex.printStackTrace(); }						
				}
			}
		}
		
		// if the corresponding bit is set at a given program point then there was no field access and no synchronization event of relevant type between the point and exit node from the method
		fieldsNumbering.add(EXIT_MARK_FIELD);
		
		markCallRetBitPos = fieldsNumbering.getMaximumIndex() + 1;
	}

	
	static class FieldsBeforeSynchEvent
	{
		private Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG;

		private CallGraph clGraph;
		
		private boolean considerReads;
		private boolean considerWrites;
		
		public FieldsBeforeSynchEvent(Graph<BasicBlockInContext<IExplodedBasicBlock>> bwICFG, CallGraph clGraph, boolean reads, boolean writes)
		{
			this.bwICFG = bwICFG;
			this.clGraph = clGraph;
			
			this.considerReads = reads;
			this.considerWrites = writes;
		}
		
		
		class TransferFunctionsFBSE implements ITransferFunctionProvider<BasicBlockInContext<IExplodedBasicBlock>, BitVectorVariable> 
		{
			public AbstractMeetOperator<BitVectorVariable> getMeetOperator() 
			{
				int markFieldNum = fieldsNumbering.getMappedIndex(EXIT_MARK_FIELD);
				
				return new BitVectorUnionFilterCR(markCallRetBitPos, markFieldNum);
			}

			public UnaryOperator<BitVectorVariable> getNodeTransferFunction(BasicBlockInContext<IExplodedBasicBlock> bb) 
			{
				IExplodedBasicBlock ebb = bb.getDelegate();
				
				SSAInstruction insn = ebb.getInstruction();
				
				if ((insn instanceof SSAInvokeInstruction) || (insn instanceof SSAReturnInstruction) || (insn instanceof SSAMonitorInstruction))
				{
					try
					{
						EnumSet<SynchEventType> eventTypes = SynchEventCodeInfo.getSynchEventTypesForInsn(bb.getNode(), bb.getNode().getIR(), insn, clGraph);
						
						if (eventTypes.contains(curEventType))
						{
							return BitVectorKillAll.getInstance();
						}
					}
					catch (Exception ex) { ex.printStackTrace(); }
					
					// other cases: synchronization events with other types, no events
					return BitVectorIdentity.instance();				
				}
				else if (insn instanceof SSAFieldAccessInstruction) 
				{
					SSAFieldAccessInstruction faInsn = (SSAFieldAccessInstruction) insn;
					
					if ((considerReads && (faInsn instanceof SSAGetInstruction)) || (considerWrites && (faInsn instanceof SSAPutInstruction)))
					{					
						BitVector kill = new BitVector();
						BitVector gen = new BitVector();
	
						Set<FieldID> fieldIDs = FieldAccessCodeInfo.getFieldsForInsn(bb.getNode(), faInsn, ebb.getFirstInstructionIndex());							
	
						try
						{
							for (FieldID fid : fieldIDs)
							{
								int fieldNum = fieldsNumbering.getMappedIndex(fid);
								
								gen.set(fieldNum);
							}
							
							// there was at least one field access
							if ( ! fieldIDs.isEmpty() )
							{
								int markFieldNum = fieldsNumbering.getMappedIndex(EXIT_MARK_FIELD);
								kill.set(markFieldNum);
							}	
						}
						catch (Exception ex) { ex.printStackTrace(); }
					
						return new BitVectorKillGen(kill, gen);
					}
					else
					{
						// identity function for all other cases
						return BitVectorIdentity.instance();	
					}
				}
				else if (ebb.isExitBlock())
				{
					BitVector kill = new BitVector();
					BitVector gen = new BitVector();
					
					int markFieldNum = fieldsNumbering.getMappedIndex(EXIT_MARK_FIELD);
					gen.set(markFieldNum);
					
					return new BitVectorKillGen(kill, gen);
				}
				else
				{
					// identity function for all other instructions
					return BitVectorIdentity.instance();
				}
			}
			
			public UnaryOperator<BitVectorVariable> getEdgeTransferFunction(BasicBlockInContext<IExplodedBasicBlock> src, BasicBlockInContext<IExplodedBasicBlock> dst) 
			{
				// we do not propagate through return-to-exit edges 
				
				if (dst.getDelegate().isExitBlock()) 
				{
					return BitVectorKillAll.getInstance();
				}
				
				// detecting call-return edges: "dst" node is invoke and "src" is not entry node
				if ( ! src.isEntryBlock() )
				{
					SSAInstruction insn = dst.getDelegate().getInstruction();
					
					if (insn instanceof SSAInvokeInstruction)
					{
						BitVector gen = new BitVector();
						BitVector kill = new BitVector();
						
						gen.set(markCallRetBitPos);
						
						return new BitVectorKillGen(kill, gen);
					}
				}
				
				// default
				return BitVectorIdentity.instance();
			}

			public boolean hasNodeTransferFunctions() 
			{
				return true;
			}
		
			public boolean hasEdgeTransferFunctions() 
			{
				return true;
			}
		}
	
		public BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> analyze() 
		{
			BitVectorFramework<BasicBlockInContext<IExplodedBasicBlock>, FieldID> framework = new BitVectorFramework<BasicBlockInContext<IExplodedBasicBlock>, FieldID>(bwICFG, new TransferFunctionsFBSE(), fieldsNumbering);
			
			BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>> solver = new BitVectorSolver<BasicBlockInContext<IExplodedBasicBlock>>(framework);

			try
			{
				solver.solve(null);
			}
			catch (Exception ex) { ex.printStackTrace(); }
			
			return solver;
		}
	}	
}

