/*
 * Copyright (C) 2015, Charles University in Prague.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package cz.cuni.mff.d3s.stdynpor.jpf;

import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;
import java.util.Stack;
import java.util.Collection;

import gov.nasa.jpf.JPF;
import gov.nasa.jpf.Config;
import gov.nasa.jpf.ListenerAdapter;
import gov.nasa.jpf.vm.VM;
import gov.nasa.jpf.vm.ClassInfo;
import gov.nasa.jpf.vm.ElementInfo;
import gov.nasa.jpf.vm.ThreadInfo;
import gov.nasa.jpf.vm.FieldInfo;
import gov.nasa.jpf.vm.Instruction;
import gov.nasa.jpf.jvm.bytecode.GETFIELD;
import gov.nasa.jpf.jvm.bytecode.GETSTATIC;
import gov.nasa.jpf.jvm.bytecode.PUTFIELD;
import gov.nasa.jpf.jvm.bytecode.PUTSTATIC;
import gov.nasa.jpf.jvm.bytecode.InstanceFieldInstruction;
import gov.nasa.jpf.jvm.bytecode.StaticFieldInstruction;
import gov.nasa.jpf.jvm.bytecode.ArrayElementInstruction;
import gov.nasa.jpf.search.Search;

import cz.cuni.mff.d3s.stdynpor.Configuration;
import cz.cuni.mff.d3s.stdynpor.data.AllocationSite;
import cz.cuni.mff.d3s.stdynpor.data.FieldID;
import cz.cuni.mff.d3s.stdynpor.data.ArrayID;
import cz.cuni.mff.d3s.stdynpor.bytecode.ByteCodeUtils;


public class SharedAccessTracker extends ListenerAdapter
{
	// current transition
	private TransitionInfo curTr;
	
	// stack of transitions (current dynamic path)
	private Stack<TransitionInfo> curPathTrs;

	private int pathLenPercents;

	public SharedAccessTracker(Config config, JPF jpf) 
	{
		curPathTrs = new Stack<TransitionInfo>();
		
		pathLenPercents = config.getInt("heuristic.dyntrace_length.percentage", 100);
	}

	public void searchStarted(Search search)
	{
		curTr = new TransitionInfo();
	}
	
	public void stateAdvanced(Search search) 
	{
		curPathTrs.push(curTr);
		
		curTr = new TransitionInfo();
	}
	
	public void stateBacktracked(Search search) 
	{
		curPathTrs.pop();
	}

	public void executeInstruction(VM vm, ThreadInfo curTh, Instruction insn)
	{
		if (curTh.isFirstStepInsn()) return;
		
		// skip accesses performed in library classes
		if ((insn instanceof InstanceFieldInstruction) || (insn instanceof ArrayElementInstruction))
		{
			if (ByteCodeUtils.isLibraryEntity(insn.getMethodInfo().getClassName())) return;
		}
			
		if (insn instanceof GETFIELD)
		{
			GETFIELD gfInsn = (GETFIELD) insn;

			FieldInfo targetFI = gfInsn.getFieldInfo();			
			
			int targetObjRef = curTh.getTopFrame().peek();
			
			if (targetObjRef == -1) return;
			
			String fieldName = targetFI.getName();
			
			AllocationSite objAS = ObjectAllocationTracker.getAllocSiteForObject(targetObjRef);
			FieldID tgtField = new FieldID(objAS, fieldName);
			
			FieldAccessInfo fa = new FieldAccessInfo(tgtField, curTh.getId(), 'r');
			
			curTr.fieldAccesses.add(fa);
		}
		
		if (insn instanceof PUTFIELD)
		{
			PUTFIELD pfInsn = (PUTFIELD) insn;

			FieldInfo targetFI = pfInsn.getFieldInfo();
			
			int targetObjRef = (targetFI.getStorageSize() == 1) ? curTh.getTopFrame().peek(1) : curTh.getTopFrame().peek(2);
			
			if (targetObjRef == -1) return;
			
			String fieldName = targetFI.getName();
			
			AllocationSite objAS = ObjectAllocationTracker.getAllocSiteForObject(targetObjRef);
			FieldID tgtField = new FieldID(objAS, fieldName);
			
			FieldAccessInfo fa = new FieldAccessInfo(tgtField, curTh.getId(), 'w');
			
			curTr.fieldAccesses.add(fa);
		}
		
		if (insn instanceof ArrayElementInstruction)
		{
			ArrayElementInstruction arrayInsn = (ArrayElementInstruction) insn;
			
			int targetArrayObjRef = arrayInsn.getArrayRef(curTh);
			
			if (targetArrayObjRef == -1) return;
			
			ElementInfo targetArrayObjEI = vm.getHeap().get(targetArrayObjRef);
			
			int arrayDimensions = JPFUtils.getArrayDimensionsCount(targetArrayObjEI);
			
			AllocationSite objAS = ObjectAllocationTracker.getAllocSiteForObject(targetArrayObjRef);
			
			// unknown object (most probably command line arguments)
			if (objAS == null) return;
			
			ArrayID tgtArray = new ArrayID(objAS, arrayDimensions);
			
			char accessType = arrayInsn.isRead() ? 'r' : 'w';

			ArrayAccessInfo aa = new ArrayAccessInfo(tgtArray, curTh.getId(), accessType);
			
			curTr.arrayAccesses.add(aa);
		}
	}

	public void instructionExecuted(VM vm, ThreadInfo curTh, Instruction nextInsn, Instruction execInsn)
	{
		if (curTh.isFirstStepInsn()) return;

		// skip accesses performed in library classes
		if (execInsn instanceof StaticFieldInstruction)
		{
			if (ByteCodeUtils.isLibraryEntity(execInsn.getMethodInfo().getClassName())) return;
		}
		
		if (execInsn instanceof GETSTATIC)
		{
			GETSTATIC gsInsn = (GETSTATIC) execInsn;

			FieldInfo targetFI = gsInsn.getFieldInfo();
			
			ClassInfo targetClass = targetFI.getClassInfo();
			
			String className = targetClass.getName();
			String fieldName = targetFI.getName();
			
			AllocationSite objAS = JPFUtils.getAllocSiteForClass(className);
			FieldID tgtField = new FieldID(objAS, fieldName);
		
			FieldAccessInfo fa = new FieldAccessInfo(tgtField, curTh.getId(), 'r');
			
			curTr.fieldAccesses.add(fa);
		} 
		
		if (execInsn instanceof PUTSTATIC)
		{
			PUTSTATIC psInsn = (PUTSTATIC) execInsn;
			
			FieldInfo targetFI = psInsn.getFieldInfo();
			
			ClassInfo targetClass = targetFI.getClassInfo();
			
			String className = targetClass.getName();
			String fieldName = targetFI.getName();
			
			AllocationSite objAS = JPFUtils.getAllocSiteForClass(className);
			FieldID tgtField = new FieldID(objAS, fieldName);
		
			FieldAccessInfo fa = new FieldAccessInfo(tgtField, curTh.getId(), 'w');
			
			curTr.fieldAccesses.add(fa);
		}
	}
	
	public Set<FieldID> getFieldsAccessedOnCurrentPath(boolean considerReads, boolean considerWrites)
	{
		Set<FieldID> fields = new HashSet<FieldID>();

		int pathLen = curPathTrs.size() + 1;

		// compute how many transitions we must process
		int pathFragmentLen = (pathLen * pathLenPercents) / 100;
		
		// process current transition
		if (pathFragmentLen > 0)
		{
			for (FieldAccessInfo fa : curTr.fieldAccesses)
			{
				if ((considerReads && (fa.accessType == 'r')) || (considerWrites && (fa.accessType == 'w')))
				{				
					fields.add(fa.tgtField);
				}
			}
			
			pathFragmentLen--;
		}
		
		// previous transitions
		for (int i = 1; i <= pathFragmentLen; i++)
		{
			// we must subtract 1 to reflect for "curTr" (see above)
			TransitionInfo tri = curPathTrs.get(pathLen - 1 - i);
			
			for (FieldAccessInfo fa : tri.fieldAccesses)
			{
				if ((considerReads && (fa.accessType == 'r')) || (considerWrites && (fa.accessType == 'w')))
				{				
					fields.add(fa.tgtField);
				}
			}
		}
		
		return fields;
	}

	public Set<ArrayID> getArraysAccessedOnCurrentPath(boolean considerReads, boolean considerWrites)
	{
		Set<ArrayID> arrays = new HashSet<ArrayID>();

		int pathLen = curPathTrs.size() + 1;

		// compute how many transitions we must process
		int pathFragmentLen = (pathLen * pathLenPercents) / 100;
		
		// process current transition
		if (pathFragmentLen > 0)
		{
			for (ArrayAccessInfo aa : curTr.arrayAccesses)
			{
				if ((considerReads && (aa.accessType == 'r')) || (considerWrites && (aa.accessType == 'w')))
				{				
					arrays.add(aa.tgtArray);
				}
			}
			
			pathFragmentLen--;
		}
		
		// previous transitions
		for (int i = 1; i <= pathFragmentLen; i++)
		{
			// we must subtract 1 to reflect for "curTr" (see above)
			TransitionInfo tri = curPathTrs.get(pathLen - 1 - i);
			
			for (ArrayAccessInfo aa : tri.arrayAccesses)
			{
				if ((considerReads && (aa.accessType == 'r')) || (considerWrites && (aa.accessType == 'w')))
				{				
					arrays.add(aa.tgtArray);
				}
			}
		}
		
		return arrays;
	}
	
	
	static class FieldAccessInfo
	{
		public FieldID tgtField;
		
		public int threadId;
		
		// values: r, w
		public char accessType;
		
		public FieldAccessInfo(FieldID fid, int threadId, char accessType)
		{
			this.tgtField = fid;
			this.threadId = threadId;
			this.accessType = accessType;
		}
	}

	static class ArrayAccessInfo
	{
		public ArrayID tgtArray;

		public int threadId;
		
		// values: r, w
		public char accessType;
		
		public ArrayAccessInfo(ArrayID aid, int threadId, char accessType)
		{
			this.tgtArray = aid;
			this.threadId = threadId;
			this.accessType = accessType;
		}
	}
	
	static class TransitionInfo
	{
		// list of field accesses in this transition (chronologically ordered)
		public List<FieldAccessInfo> fieldAccesses;

		// list of array accesses in this transition (chronologically ordered)
		public List<ArrayAccessInfo> arrayAccesses;
		
		public TransitionInfo()
		{
			fieldAccesses = new ArrayList<FieldAccessInfo>();
			arrayAccesses = new ArrayList<ArrayAccessInfo>();
		}
	}

}
