/*
 * Copyright (C) 2015, Charles University in Prague.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package cz.cuni.mff.d3s.stdynpor.jpf;

import java.util.Set;
import java.util.List;

import gov.nasa.jpf.Config;
import gov.nasa.jpf.vm.VM;
import gov.nasa.jpf.vm.SystemState;
import gov.nasa.jpf.vm.ThreadInfo;
import gov.nasa.jpf.vm.ElementInfo;
import gov.nasa.jpf.vm.MethodInfo;
import gov.nasa.jpf.vm.Instruction;
import gov.nasa.jpf.vm.DefaultSchedulerFactory;
import gov.nasa.jpf.vm.ChoiceGenerator;
import gov.nasa.jpf.vm.choice.ThreadChoiceFromSet;
import gov.nasa.jpf.util.StringSetMatcher;

import cz.cuni.mff.d3s.stdynpor.Configuration;
import cz.cuni.mff.d3s.stdynpor.AnalysisResultProcessor;
import cz.cuni.mff.d3s.stdynpor.Debug;
import cz.cuni.mff.d3s.stdynpor.data.ProgramPoint;
import cz.cuni.mff.d3s.stdynpor.data.FieldID;
import cz.cuni.mff.d3s.stdynpor.data.ArrayID;
import cz.cuni.mff.d3s.stdynpor.bytecode.ByteCodeUtils;


/**
 * Supports: reordering transitions (enabled threads) at choices based on the hybrid analysis results, pruning transitions based on hybrid analysis.
 */
public class TransitionSchedulerFactory extends ArraySchedulerFactory 
{
	// true: reordering transitions based on hybrid analyses
	private boolean enabledReordering;
	
	// true: pruning transitions based on hybrid analyses
	private boolean enabledPruning;
	
	// true: we distinguish between read and write accesses
	// false: default
	private boolean precisionReadWrite;
	
	public static boolean printResultsFull = false;
	public static boolean printChoiceFilter = false;
	
	
	public TransitionSchedulerFactory(Config config, VM vm, SystemState ss)
	{
		super(config, vm, ss);
		
		enabledReordering = config.getBoolean("heuristic.transitions.reordering", false);
		enabledPruning = config.getBoolean("heuristic.transitions.pruning", false);
		
		precisionReadWrite = config.getBoolean("heuristic.precision.read_write", false);
	}
	
	protected ThreadInfo[] filter(ThreadInfo[] inThList)
	{
		// we want to change order of enabled threads (transitions)
		// 1) get fields and arrays that were already accessed on the current dynamic path (some fragment given by percentage)
		// 2) identify all threads that may possibly access the same fields and arrays in future during the program execution
		// 3) move all the conflicting threads to the beginning of the list (give them priority over the other enabled threads)

		// we also want to eliminate unnecessary transitions
		// idea: make choices only for threads that have possible future conflicting accesses
		
		ThreadInfo curTh = ThreadInfo.getCurrentThread();
		Instruction curInsn = curTh.getPC();

		// there cannot be any reordering and pruning when the input array of ThreadInfo objects has only a single element
		if (inThList.length == 1) return inThList;
		
		if (ByteCodeUtils.isLibraryEntity(curInsn.getMethodInfo().getClassName())) return inThList;
		
		if (printChoiceFilter)
		{
			String curMethodSig = JPFUtils.getMethodSignature(curInsn.getMethodInfo());
			
			System.out.print("[CHOICE FILTER] new request: curMethodSig = " + curMethodSig + ", curInsnPos = " + curInsn.getPosition());			
			System.out.print(", input thread list = ");
			for (ThreadInfo th : inThList) System.out.print(th.getId() + " ");			
			System.out.println("");
		}

		if (printChoiceFilter)
		{
			for (ThreadInfo th : inThList)
			{
				List<ProgramPoint> thPPs = JPFUtils.getProgramPointsForThreadStack(th, Configuration.allowExcludingChildThreads);
				ProgramPoint thTopPP = JPFUtils.getProgramPointForThreadPC(th, Configuration.allowExcludingChildThreads);

				if (thTopPP != ProgramPoint.INVALID)
				{
					System.out.println("[CHOICE FILTER] thread = " + th.getId() + ", methodSig = " + thTopPP.methodSig + ", insnPos = " + thTopPP.insnPos + ", call stack = " + thPPs);
				}
				else
				{
					System.out.println("[CHOICE FILTER] thread = " + th.getId() + ", empty stack of program points");
				}
			}
		}
		
		SharedAccessTracker accessTracker = vm.getNextListenerOfType(SharedAccessTracker.class, null);
		
		Set<FieldID> fieldsPastRead = null;
		Set<FieldID> fieldsPastWrite = null;
		
		if (precisionReadWrite)
		{
			fieldsPastRead = accessTracker.getFieldsAccessedOnCurrentPath(true, false);
			fieldsPastWrite = accessTracker.getFieldsAccessedOnCurrentPath(false, true);
		}
		else
		{
			// if the read and write accesses are not distinguished then both sets are identical
			Set<FieldID> fieldsPastAccess = accessTracker.getFieldsAccessedOnCurrentPath(true, true);
			fieldsPastRead = fieldsPastAccess;
			fieldsPastWrite = fieldsPastAccess;
		}
		
		if (printResultsFull)
		{
			System.out.println("[FULL RESULTS] past field read accesses on the current path");			
			for (FieldID fid : fieldsPastRead)
			{
				if ( ! Debug.isWatchedEntity(fid.tgtObjectID.toString()) ) continue;
				System.out.println("\t " + fid.toString());
			}
			
			System.out.println("[FULL RESULTS] past field write accesses on the current path");
			for (FieldID fid : fieldsPastWrite)
			{
				if ( ! Debug.isWatchedEntity(fid.tgtObjectID.toString()) ) continue;
				System.out.println("\t " + fid.toString());
			}
		}
		
		Set<ArrayID> arraysPastRead = null;
		Set<ArrayID> arraysPastWrite = null;
		
		if (precisionReadWrite)
		{
			arraysPastRead = accessTracker.getArraysAccessedOnCurrentPath(true, false);
			arraysPastWrite = accessTracker.getArraysAccessedOnCurrentPath(false, true);
		}
		else
		{
			// if the read and write accesses are not distinguished then both sets are identical
			Set<ArrayID> arraysPastAccess = accessTracker.getArraysAccessedOnCurrentPath(true, true);
			arraysPastRead = arraysPastAccess;
			arraysPastWrite = arraysPastAccess;
		}
		
		if (printResultsFull)
		{
			System.out.println("[FULL RESULTS] past array read accesses on the current path");
			for (ArrayID aid : arraysPastRead)
			{
				if ( ! Debug.isWatchedEntity(aid.heapObjectID.toString()) ) continue;
				System.out.println("\t " + aid.toString());
			}
				
			System.out.println("[FULL RESULTS] past array write accesses on the current path");			
			for (ArrayID aid : arraysPastWrite)
			{
				if ( ! Debug.isWatchedEntity(aid.heapObjectID.toString()) ) continue;
				System.out.println("\t " + aid.toString());
			}
		}
		
		List<ThreadInfo> conflictingThreadsFields = AnalysisResultProcessor.getThreadsPerformingConflictFieldAccess(inThList, fieldsPastRead, fieldsPastWrite, Configuration.allowExcludingChildThreads);

		if (printChoiceFilter)
		{
			System.out.print("[CHOICE FILTER] conflicting threads (fields) = ");			
			for (ThreadInfo th : conflictingThreadsFields) System.out.print(th.getId() + " ");			
			System.out.println("");
		}
		
		List<ThreadInfo> conflictingThreadsArrays = AnalysisResultProcessor.getThreadsPerformingConflictArrayAccess(inThList, arraysPastRead, arraysPastWrite, Configuration.allowExcludingChildThreads);
		
		if (printChoiceFilter)
		{
			System.out.print("[CHOICE FILTER] conflicting threads (arrays) = ");			
			for (ThreadInfo th : conflictingThreadsArrays) System.out.print(th.getId() + " ");			
			System.out.println("");
		}
		
		ThreadInfo[] conflictThreads = new ThreadInfo[inThList.length];
		int conflictThNum = 0;
		
		ThreadInfo[] otherThreads = new ThreadInfo[inThList.length];
		int otherThNum = 0;
		
		// we must preserve the original order in each group
		for (int i = 0; i < inThList.length; i++)
		{
			ThreadInfo th = inThList[i];
			
			if (conflictingThreadsFields.contains(th) || conflictingThreadsArrays.contains(th))
			{
				conflictThreads[conflictThNum++] = th;
			}
			else
			{
				otherThreads[otherThNum++] = th;
			}
		}

		// input thread list is the default value for the case of disabled reordering and pruning
		ThreadInfo[] newThList = inThList;
		
		if (enabledReordering)
		{
			newThList = new ThreadInfo[inThList.length];
			int newThPos = 0;

			System.arraycopy(conflictThreads, 0, newThList, newThPos, conflictThNum);		
			newThPos += conflictThNum;
		
			System.arraycopy(otherThreads, 0, newThList, newThPos, otherThNum);
			newThPos += otherThNum;
		}
		
		if (enabledPruning)
		{
			if (conflictThNum > 0)
			{			
				newThList = new ThreadInfo[conflictThNum];
				int newThPos = 0;
			
				System.arraycopy(conflictThreads, 0, newThList, newThPos, conflictThNum);		
				newThPos += conflictThNum;
			}
			else
			{
				// use the first thread in the input list to allow execution continue
				newThList = new ThreadInfo[1];
				newThList[0] = inThList[0];
			}
		}
		
		if (printChoiceFilter)
		{
			System.out.print("[CHOICE FILTER] result thread list = ");
			for (ThreadInfo th : newThList) System.out.print(th.getId() + " ");
			System.out.println("");

			// blank line
			System.out.println("");
		}
		
		return newThList;
	}
}
