/*
 * Copyright (C) 2015, Charles University in Prague.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package cz.cuni.mff.d3s.stdynpor.jpf.tools;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Set;
import java.util.HashSet;
import java.util.Stack;

import gov.nasa.jpf.JPF;
import gov.nasa.jpf.Config;
import gov.nasa.jpf.ListenerAdapter;
import gov.nasa.jpf.vm.VM;
import gov.nasa.jpf.vm.ThreadInfo;
import gov.nasa.jpf.vm.ElementInfo;
import gov.nasa.jpf.vm.ClassInfo;
import gov.nasa.jpf.vm.FieldInfo;
import gov.nasa.jpf.vm.Instruction;
import gov.nasa.jpf.jvm.bytecode.GETFIELD;
import gov.nasa.jpf.jvm.bytecode.GETSTATIC;
import gov.nasa.jpf.jvm.bytecode.PUTFIELD;
import gov.nasa.jpf.jvm.bytecode.PUTSTATIC;
import gov.nasa.jpf.search.Search;


public class FieldAccessTracker extends ListenerAdapter 
{
	// current transition
	private TransitionInfo curTr;
	
	// stack of transitions (current path)
	private Stack<TransitionInfo> curPathTrs;

	// all field accesses that happened on the current state space path
	private LinkedList<FieldAccessInfo> curPathAccesses;

	// observed conflicts
	private Set<String> conflictFieldAccesses;


	public FieldAccessTracker(Config config, JPF jpf) 
	{
		curPathTrs = new Stack<TransitionInfo>();
		
		curPathAccesses = new LinkedList<FieldAccessInfo>();
		
		conflictFieldAccesses = new HashSet<String>();
    }


	public void searchStarted(Search search)
	{
		curTr = new TransitionInfo();
	}
	
	public void stateAdvanced(Search search) 
	{
		for (int i = 0; i < curTr.fieldAccesses.size(); i++)
		{
			FieldAccessInfo curFA = curTr.fieldAccesses.get(i);
			
			// check whether this field was accessed also in the past in a conflicting way
			
			boolean existsPastFA = false;
			FieldAccessInfo pastFA = null;
			
			for (FieldAccessInfo fa : curPathAccesses)
			{
				if (fa.className.equals(curFA.className) && fa.fieldName.equals(curFA.fieldName) && (fa.threadId != curFA.threadId) && (fa.accessType != curFA.accessType))
				{
					existsPastFA = true;
					pastFA = fa;
					break;
				}
			}
			if (existsPastFA)
			{
				String description = "field name = " + curFA.className + "." + curFA.fieldName + ", current location = " + curFA.accessLocation + " (th = " + curFA.threadId + "), past location = " + pastFA.accessLocation + " (th = " + pastFA.threadId + ")";
				
				if ( ! conflictFieldAccesses.contains(description) )
				{
					conflictFieldAccesses.add(description);
					System.out.println("[CONFLICT FIELD ACCESS] " + description);
				}
			}
			
			curPathAccesses.addLast(curFA);
		}
		
		curPathTrs.push(curTr);
		
		curTr = new TransitionInfo();
	}
	
	public void stateBacktracked(Search search) 
	{
		TransitionInfo tri = curPathTrs.peek();
		
		for (int i = 0; i < tri.fieldAccesses.size(); i++)
		{
			curPathAccesses.removeLast();	
		}
	
		curPathTrs.pop();
	}
	
	public void executeInstruction(VM vm, ThreadInfo curTh, Instruction insn)
	{
		if (curTh.isFirstStepInsn()) return;
	
		if (insn instanceof GETFIELD)
		{
			GETFIELD gfInsn = (GETFIELD) insn;

			FieldInfo targetFI = gfInsn.getFieldInfo();			
			
			int targetObjRef = curTh.getTopFrame().peek();
			ElementInfo targetObjEI = vm.getHeap().get(targetObjRef);
			
			String className = targetObjEI.getClassInfo().getName();
			String fieldName = targetFI.getName();
		
			String accessLoc = insn.getMethodInfo().getBaseName() + insn.getMethodInfo().getSignature() + ":" + String.valueOf(insn.getPosition());
			
			FieldAccessInfo fa = new FieldAccessInfo(className, fieldName, curTh.getId(), accessLoc, 'r');
			
			curTr.fieldAccesses.add(fa);
		}
		
		if (insn instanceof PUTFIELD)
		{
			PUTFIELD pfInsn = (PUTFIELD) insn;

			FieldInfo targetFI = pfInsn.getFieldInfo();
			
			int targetObjRef = (targetFI.getStorageSize() == 1) ? curTh.getTopFrame().peek(1) : curTh.getTopFrame().peek(2);	
			ElementInfo targetObjEI = vm.getHeap().get(targetObjRef);
			
			String className = targetObjEI.getClassInfo().getName();
			String fieldName = targetFI.getName();
		
			String accessLoc = insn.getMethodInfo().getBaseName() + insn.getMethodInfo().getSignature() + ":" + String.valueOf(insn.getPosition());
			
			FieldAccessInfo fa = new FieldAccessInfo(className, fieldName, curTh.getId(), accessLoc, 'w');
			
			curTr.fieldAccesses.add(fa);
		}	
	}

	public void instructionExecuted(VM vm, ThreadInfo curTh, Instruction nextInsn, Instruction execInsn)
	{
		if (curTh.isFirstStepInsn()) return;

		if (execInsn instanceof GETSTATIC)
		{
			GETSTATIC gsInsn = (GETSTATIC) execInsn;

			FieldInfo targetFI = gsInsn.getFieldInfo();
			
			ClassInfo targetClass = targetFI.getClassInfo();
			
			String className = targetClass.getName();
			String fieldName = targetFI.getName();
		
			String accessLoc = gsInsn.getMethodInfo().getBaseName() + gsInsn.getMethodInfo().getSignature() + ":" + String.valueOf(gsInsn.getPosition());
			
			FieldAccessInfo fa = new FieldAccessInfo(className, fieldName, curTh.getId(), accessLoc, 'r');
			
			curTr.fieldAccesses.add(fa);
		} 
		
		if (execInsn instanceof PUTSTATIC)
		{
			PUTSTATIC psInsn = (PUTSTATIC) execInsn;
			
			FieldInfo targetFI = psInsn.getFieldInfo();
			
			ClassInfo targetClass = targetFI.getClassInfo();
			
			String className = targetClass.getName();
			String fieldName = targetFI.getName();
		
			String accessLoc = psInsn.getMethodInfo().getBaseName() + psInsn.getMethodInfo().getSignature() + ":" + String.valueOf(psInsn.getPosition());
			
			FieldAccessInfo fa = new FieldAccessInfo(className, fieldName, curTh.getId(), accessLoc, 'w');
			
			curTr.fieldAccesses.add(fa);
		}
	}
    

	static class FieldAccessInfo
	{
		public String className;
		public String fieldName;
		
		public int threadId;
		
		// method signature and bytecode position
		public String accessLocation;
		
		public char accessType;
		
		public FieldAccessInfo(String cn, String fn, int tid, String loc, char at)
		{
			this.className = cn;
			this.fieldName = fn;
			this.threadId = tid;
			this.accessLocation = loc;
			this.accessType = at;
		}
	}
	
    static class TransitionInfo
	{
		// list of field accesses in this transition (chronologically ordered)
		public List<FieldAccessInfo> fieldAccesses;

		public TransitionInfo()
		{
			fieldAccesses = new ArrayList<FieldAccessInfo>();
		}
	}

}
