/*
 * Copyright (C) 2015, Charles University in Prague.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package cz.cuni.mff.d3s.stdynpor.wala;

import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.EnumSet;

import com.ibm.wala.ipa.callgraph.CallGraph;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.types.TypeReference;
import com.ibm.wala.ssa.SSAOptions;
import com.ibm.wala.ssa.IR;
import com.ibm.wala.ssa.DefUse;
import com.ibm.wala.ssa.SSAInstruction;
import com.ibm.wala.ssa.SSAInvokeInstruction;
import com.ibm.wala.ssa.SSAReturnInstruction;
import com.ibm.wala.ssa.SSAMonitorInstruction;
import com.ibm.wala.ssa.SSALoadMetadataInstruction;
import com.ibm.wala.analysis.typeInference.TypeInference;

import cz.cuni.mff.d3s.stdynpor.Configuration;
import cz.cuni.mff.d3s.stdynpor.data.ProgramPoint;
import cz.cuni.mff.d3s.stdynpor.data.AllocationSite;
import cz.cuni.mff.d3s.stdynpor.data.ClassName;
import cz.cuni.mff.d3s.stdynpor.data.SynchEventID;
import cz.cuni.mff.d3s.stdynpor.data.SynchEventType;
import cz.cuni.mff.d3s.stdynpor.bytecode.ByteCodeUtils;


public class SynchEventCodeInfo
{
	public static Set<SynchEventID> getSynchEventsForInsn(CGNode mthNode, IR methodIR, SSAInstruction insn, CallGraph clGraph) throws Exception
	{
		// synchronization events represented by the current instruction
		Set<SynchEventID> events = new HashSet<SynchEventID>();
		
		// we ignore synchronization events inside Object.wait() and Thread.join()
		String curMethodSig = WALAUtils.getMethodSignature(mthNode.getMethod());
		if (ByteCodeUtils.isSynchEventMethod(curMethodSig)) return events;			
					
		if (insn instanceof SSAInvokeInstruction)
		{
			SSAInvokeInstruction invokeInsn = (SSAInvokeInstruction) insn;
			
			String targetMthName = WALAUtils.getShortMethodName(invokeInsn.getDeclaredTarget());
			
			List<IMethod> targetMths = WALAUtils.findTargetMethods(mthNode, invokeInsn, clGraph);
			
			SynchEventType evType = null;
			if (targetMthName.equals("wait")) evType = SynchEventType.WAIT;
			if (targetMthName.equals("notify") || targetMthName.equals("notifyAll")) evType = SynchEventType.NOTIFY;
			if (targetMthName.equals("start"))
			{
				if (WALAUtils.isThreadClass(invokeInsn.getDeclaredTarget().getDeclaringClass())) evType = SynchEventType.START;
			}
			if (targetMthName.equals("join"))
			{
				if (WALAUtils.isThreadClass(invokeInsn.getDeclaredTarget().getDeclaringClass())) evType = SynchEventType.JOIN;
			}
			
			if (evType != null)
			{
				// we do not have to check for static methods here because none of the relevant methods is (Object.wait, Object.notify, Thread.start, Thread.join) -> there must be a receiver object
				
				String callerMethodSig = WALAUtils.getMethodSignature(mthNode.getMethod());
			
				String rcvClassName = WALAUtils.getDeclaringClassNameForMethod(invokeInsn.getDeclaredTarget());

				Set<AllocationSite> rcvObjectAllocSites = PointerAnalysisData.getObjectAllocSites(rcvClassName, callerMethodSig, invokeInsn.getReceiver());
	
				for (AllocationSite objAS : rcvObjectAllocSites) events.add(new SynchEventID(objAS, evType));
			}
			
			if (targetMths != null)
			{
				for (IMethod tgtMth : targetMths)
				{			
					if (tgtMth.isSynchronized())
					{
						String tgtMthClassName = WALAUtils.getDeclaringClassNameForMethod(tgtMth);
						
						if (tgtMth.isStatic())
						{								
							events.add(new SynchEventID(WALAUtils.getAllocationSiteForClass(tgtMthClassName), SynchEventType.LOCK));
						}
						else
						{							
							String callerMethodSig = WALAUtils.getMethodSignature(mthNode.getMethod());
							
							Set<AllocationSite> rcvObjectAllocSites = PointerAnalysisData.getObjectAllocSites(tgtMthClassName, callerMethodSig, invokeInsn.getReceiver());
			
							for (AllocationSite objAS : rcvObjectAllocSites) events.add(new SynchEventID(objAS, SynchEventType.LOCK));
						}
					}
				}
			}
		}
		
		if (insn instanceof SSAReturnInstruction)
		{	
			IMethod mth = mthNode.getMethod();
			
			if (mth.isSynchronized())
			{
				String methodSig = WALAUtils.getMethodSignature(mth);
				
				String methodClassName = WALAUtils.getDeclaringClassNameForMethod(mth);
				
				if (mth.isStatic())
				{								
					events.add(new SynchEventID(WALAUtils.getAllocationSiteForClass(methodClassName), SynchEventType.UNLOCK));
				}
				else
				{				
					Set<AllocationSite> thisObjectAllocSites = PointerAnalysisData.getObjectAllocSites(methodClassName, methodSig, 1);
			
					for (AllocationSite objAS : thisObjectAllocSites) events.add(new SynchEventID(objAS, SynchEventType.UNLOCK));
				}
			}
		}
		
		if (insn instanceof SSAMonitorInstruction)
		{
			SSAMonitorInstruction monInsn = (SSAMonitorInstruction) insn;

			String methodSig = WALAUtils.getMethodSignature(mthNode.getMethod());
			
			SynchEventType evType = null;
			if (monInsn.isMonitorEnter()) evType = SynchEventType.LOCK;
			else evType = SynchEventType.UNLOCK;
					
			TypeInference mthVarTypeInfo = TypeAnalysisData.getMethodTypeInfo(methodSig, methodIR); 
			
			TypeReference tgtObjTypeRef = mthVarTypeInfo.getType(monInsn.getRef()).getTypeReference(); 
			String tgtObjClassName = WALAUtils.getClassName(tgtObjTypeRef);
			
			// process special case: "synchronized (<class name>.class)"
			if (tgtObjClassName.equals("java.lang.Class"))
			{
				// probably expensive but used rarely
				DefUse mthDU = new DefUse(methodIR);
				
				SSAInstruction tgtObjDefInsn = mthDU.getDef(monInsn.getRef());
						
				// def insn for monInsn.getRef() should be ldc_w
				if (tgtObjDefInsn instanceof SSALoadMetadataInstruction)
				{
					SSALoadMetadataInstruction metaInsn = (SSALoadMetadataInstruction) tgtObjDefInsn;
					
					// get class name from the constant pool
					String lvarClassName = WALAUtils.getClassName((TypeReference) metaInsn.getToken());
					
					events.add(new SynchEventID(WALAUtils.getAllocationSiteForClass(lvarClassName), evType));
				}
			}
			
			Set<AllocationSite> tgtObjectAllocSites = PointerAnalysisData.getObjectAllocSites(tgtObjClassName, methodSig, monInsn.getRef());
							
			for (AllocationSite objAS : tgtObjectAllocSites) events.add(new SynchEventID(objAS, evType));
		}
		
		return events;
	}	

	public static Set<SynchEventID> getUnblockEventsForInsn(CGNode mthNode, IR methodIR, SSAInstruction insn) throws Exception
	{
		Set<SynchEventID> events = new HashSet<SynchEventID>();
		
		// we ignore synchronization events inside Object.wait() and Thread.join()
		String curMethodSig = WALAUtils.getMethodSignature(mthNode.getMethod());
		if (ByteCodeUtils.isSynchEventMethod(curMethodSig)) return events;
		
		if (insn instanceof SSAInvokeInstruction)
		{
			SSAInvokeInstruction invokeInsn = (SSAInvokeInstruction) insn;
			
			String targetMthName = WALAUtils.getShortMethodName(invokeInsn.getDeclaredTarget());
			
			if (targetMthName.equals("notify") || targetMthName.equals("notifyAll"))
			{
				// we do not have to check for static methods here because none of the relevant methods is (Object.notify, Object.notifyAll) -> there must be a receiver object
				
				String callerMethodSig = WALAUtils.getMethodSignature(mthNode.getMethod());
			
				String rcvClassName = WALAUtils.getDeclaringClassNameForMethod(invokeInsn.getDeclaredTarget());

				Set<AllocationSite> rcvObjectAllocSites = PointerAnalysisData.getObjectAllocSites(rcvClassName, callerMethodSig, invokeInsn.getReceiver());
	
				for (AllocationSite objAS : rcvObjectAllocSites) events.add(new SynchEventID(objAS, SynchEventType.NOTIFY));
			}
		}
		
		if (insn instanceof SSAReturnInstruction)
		{	
			IMethod mth = mthNode.getMethod();
			
			if (mth.isSynchronized())
			{
				String methodSig = WALAUtils.getMethodSignature(mth);
				
				String methodClassName = WALAUtils.getDeclaringClassNameForMethod(mth);
			
				if (mth.isStatic())
				{								
					events.add(new SynchEventID(WALAUtils.getAllocationSiteForClass(methodClassName), SynchEventType.UNLOCK));
				}
				else
				{
					Set<AllocationSite> thisObjectAllocSites = PointerAnalysisData.getObjectAllocSites(methodClassName, methodSig, 1);
			
					for (AllocationSite objAS : thisObjectAllocSites) events.add(new SynchEventID(objAS, SynchEventType.UNLOCK));
				}
			}
		}
		
		if (insn instanceof SSAMonitorInstruction)
		{				
			SSAMonitorInstruction monInsn = (SSAMonitorInstruction) insn;
			
			// we care only about "unlock" events here
			if ( ! monInsn.isMonitorEnter() )
			{	
				String methodSig = WALAUtils.getMethodSignature(mthNode.getMethod());
				
				TypeInference mthVarTypeInfo = TypeAnalysisData.getMethodTypeInfo(methodSig, methodIR);
				
				TypeReference tgtObjTypeRef = mthVarTypeInfo.getType(monInsn.getRef()).getTypeReference(); 
				String tgtObjClassName = WALAUtils.getClassName(tgtObjTypeRef);
				
				// process special case: "synchronized (<class name>.class)"
				if (tgtObjClassName.equals("java.lang.Class"))
				{
					// probably expensive but used rarely					
					DefUse mthDU = new DefUse(methodIR);
					
					SSAInstruction tgtObjDefInsn = mthDU.getDef(monInsn.getRef());
							
					// def insn for monInsn.getRef() should be ldc_w
					if (tgtObjDefInsn instanceof SSALoadMetadataInstruction)
					{						
						SSALoadMetadataInstruction metaInsn = (SSALoadMetadataInstruction) tgtObjDefInsn;
						
						// get class name from the constant pool
						String lvarClassName = WALAUtils.getClassName((TypeReference) metaInsn.getToken());
						
						events.add(new SynchEventID(WALAUtils.getAllocationSiteForClass(lvarClassName), SynchEventType.UNLOCK));
					}
				}
				
				Set<AllocationSite> tgtObjectAllocSites = PointerAnalysisData.getObjectAllocSites(tgtObjClassName, methodSig, monInsn.getRef());
			
				for (AllocationSite objAS : tgtObjectAllocSites) events.add(new SynchEventID(objAS, SynchEventType.UNLOCK));
			}
		}
		
		return events;
	}
	
	public static EnumSet<SynchEventType> getSynchEventTypesForInsn(CGNode mthNode, IR methodIR, SSAInstruction insn, CallGraph clGraph) throws Exception
	{
		EnumSet<SynchEventType> eventTypes = EnumSet.noneOf(SynchEventType.class);
		
		// we ignore synchronization events inside Object.wait() and Thread.join()
		String curMethodSig = WALAUtils.getMethodSignature(mthNode.getMethod());
		if (ByteCodeUtils.isSynchEventMethod(curMethodSig)) return eventTypes;			
		
		if (insn instanceof SSAInvokeInstruction)
		{
			SSAInvokeInstruction invokeInsn = (SSAInvokeInstruction) insn;
			
			String targetMthName = WALAUtils.getShortMethodName(invokeInsn.getDeclaredTarget());
			
			List<IMethod> targetMths = WALAUtils.findTargetMethods(mthNode, invokeInsn, clGraph);
			
			if (targetMthName.equals("wait")) eventTypes.add(SynchEventType.WAIT);
			if (targetMthName.equals("notify") || targetMthName.equals("notifyAll")) eventTypes.add(SynchEventType.NOTIFY);
			if (targetMthName.equals("start"))
			{
				if (WALAUtils.isThreadClass(invokeInsn.getDeclaredTarget().getDeclaringClass())) eventTypes.add(SynchEventType.START);
			}
			if (targetMthName.equals("join"))
			{
				if (WALAUtils.isThreadClass(invokeInsn.getDeclaredTarget().getDeclaringClass())) eventTypes.add(SynchEventType.JOIN);
			}
		
			if (targetMths != null)
			{
				for (IMethod tgtMth : targetMths)
				{			
					if (tgtMth.isSynchronized()) eventTypes.add(SynchEventType.LOCK);
				}
			}
		}
		
		if (insn instanceof SSAReturnInstruction)
		{	
			IMethod mth = mthNode.getMethod();
			
			if (mth.isSynchronized()) eventTypes.add(SynchEventType.UNLOCK);
		}
		
		if (insn instanceof SSAMonitorInstruction)
		{
			SSAMonitorInstruction monInsn = (SSAMonitorInstruction) insn;

			if (monInsn.isMonitorEnter()) eventTypes.add(SynchEventType.LOCK);
			else eventTypes.add(SynchEventType.UNLOCK);		
		}

		return eventTypes;
	}

}

